from enum import IntEnum


class ScenarioVariant(IntEnum):
    """
    This enum class provides the integer values used to reference the different variants of scenarios (and the game).
    """
    LEGACY = 0
    AOE2 = 1
    ROR = 2

    def to_display_name(self) -> str:
        return {
            ScenarioVariant.LEGACY: 'Legacy',
            ScenarioVariant.AOE2: 'Age of Empires 2',
            ScenarioVariant.ROR:  'Return of Rome',
        }[self]

    def applicants(self):
        return {
            ScenarioVariant.LEGACY: 'AoC & Bare HD',
            ScenarioVariant.AOE2: 'HD with expansions & DE',
            ScenarioVariant.ROR:  'Return of Rome',
        }[self]



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional
from uuid import UUID
from weakref import WeakValueDictionary

from AoE2ScenarioParser.objects.support.uuid_list import NO_UUID

if TYPE_CHECKING:
    from AoE2ScenarioParser.scenarios.aoe2_scenario import AoE2Scenario
    from AoE2ScenarioParser.objects.aoe2_object import AoE2Object

_scenarios: WeakValueDictionary[UUID, 'AoE2Scenario'] = WeakValueDictionary()
_scenarios_by_name: WeakValueDictionary[str, 'AoE2Scenario'] = WeakValueDictionary()


def get_scenario(
        uuid: UUID = None,
        obj: 'AoE2Object' = None,
        name: str = None
) -> Optional['AoE2Scenario']:
    """
    Get scenario through a UUID, a related object or the name of a scenario.
    Not intended to be called outside the store itself.

    Use the wrapper instead: `AoE2Scenario.get_scenario()`

    Args:
        uuid: The UUID of the scenario
        obj: An object related to a scenario
        name: The name of a scenario

    Returns:
        The scenario based on the given identifier, or `None`
    """
    if uuid == NO_UUID:
        return None

    key, dict_ = "", {}
    if uuid is not None:
        key, dict_ = uuid, _scenarios
    elif obj is not None:
        key, dict_ = obj._uuid, _scenarios
    elif name is not None:
        key, dict_ = name, _scenarios_by_name

    if key in dict_:
        return dict_[key]
    else:
        raise ValueError("Unable to find scenario based on the given identifier")


def register_scenario(scenario: 'AoE2Scenario') -> None:
    """
    Register a scenario to the store

    Args:
        scenario: The scenario to register
    """
    if scenario.uuid in _scenarios:
        raise ValueError("Scenario with that UUID already present")
    _scenarios[scenario.uuid] = scenario
    _scenarios_by_name[scenario.name] = scenario


def remove_scenario(uuid: UUID) -> None:
    """
    force remove a scenario from the store. In general this is not necessary as the store stores references to the
    scenario weakly. This means, as soon as the user removes all hard references, it's gone from the store too.

    Args:
        uuid: The UUID of the scenario
    """
    del _scenarios[uuid]



from __future__ import annotations

from enum import Enum
from typing import List


class InfoDatasetBase(Enum):
    """
    This enum class is the base class for the UnitInfo, BuildingInfo, HeroInfo and OtherInfo datasets. It provides the
    properties and functions common to all the mentioned classes

    **Methods**
    
    >>> InfoDatasetBase.from_id()
    >>> InfoDatasetBase.from_dead_id()
    >>> InfoDatasetBase.from_icon_id()
    >>> InfoDatasetBase.from_hotkey_id()
    >>> InfoDatasetBase.gaia_only()
    >>> InfoDatasetBase.non_gaia()
    """

    @staticmethod
    def _id_map() -> dict:
        """
        Returns:
            A dict that maps the names of the properties to their index in the info tuple
        """
        return {'id': 0, 'icon_id': 1, 'dead_id': 2, 'hotkey_id': 3, 'gaia_only': 4}

    def _get_property(self, name: str) -> int | bool:
        """
        Get the specified property by its name from an object of this (or its derived) class

        Args:
            name: the name of the property to get

        Returns:
        """
        return self.value[self._id_map()[name]]

    @property
    def ID(self) -> int:
        """
        Returns:
            The ID of the specified unit
        """
        return self._get_property('id')

    @property
    def ICON_ID(self) -> int:
        """
        Returns:
            The Icon ID of the specified unit
        """
        return self._get_property('icon_id')

    @property
    def DEAD_ID(self) -> int:
        """
        Returns:
            The Dead Unit ID of the specified unit
        """
        return self._get_property('dead_id')

    @property
    def HOTKEY_ID(self) -> int:
        """
        Returns:
            The HotKey ID of the specified unit
        """
        return self._get_property('hotkey_id')

    @property
    def IS_GAIA_ONLY(self) -> bool:
        """
        Returns:
            A boolean value indicating if the specified unit is a gaia only unit (eg. Deer)
        """
        return self._get_property('gaia_only')

    @classmethod
    def _from_id(cls, id_type: str, value: int) -> InfoDatasetBase:
        """
        Finds and returns the member object that uses the given value for the specified property (id_type)

        Args:
            id_type: the property of member objects that should match the value specified
            value: the value of the property to search for

        Returns:
            An InfoDatasetBase member object which uses the given value for the specified property (id_type
        """
        index = cls._id_map()[id_type]

        if type(value) is not int:
            raise TypeError(f"from_id expected int, got {type(value)}")
        if value < 0:
            raise ValueError(f"{value} is not a valid id value")

        for member in cls._member_map_.values():
            if member.value[index] == value:
                return member

        raise KeyError(f"A unit with {id_type} = {value} was not found in the dataset")

    @classmethod
    def from_id(cls, unit_id: int) -> InfoDatasetBase:
        """
        Finds and returns the unit with the given unit ID

        Args:
            unit_id: the unit ID to search for

        Returns:
            A unit with the given unit I
        """
        return cls._from_id('id', unit_id)

    @classmethod
    def from_icon_id(cls, icon_id: int) -> InfoDatasetBase:
        """
        Finds and returns the unit with the given icon ID

        Args:
            icon_id: the icon ID to search for

        Returns:
            A unit with the given icon ID
        """
        return cls._from_id('icon_id', icon_id)

    @classmethod
    def from_dead_id(cls, dead_id: int) -> InfoDatasetBase:
        """
        Finds and returns the unit with the given dead unit ID

        Args:
            dead_id: the dead unit ID to search for

        Returns:
            A unit with the given dead unit ID
        """
        return cls._from_id('dead_id', dead_id)

    @classmethod
    def from_hotkey_id(cls, hotkey_id: int) -> InfoDatasetBase:
        """
        Finds and returns the unit with the given hotkey ID. Note that there may be multiple units that
        use the same hotkey ID, currently only one is returned!

        Args:
            hotkey_id: the hotkey ID to search for

        Returns:
            A unit with the given hotkey ID
        """
        return cls._from_id('hotkey_id', hotkey_id)

    @classmethod
    def gaia_only(cls) -> List[InfoDatasetBase]:
        """
        Returns:
            A list of all the gaia only units (eg. Deer)
        """
        return cls._gaia_filter(gaia_only=True)

    @classmethod
    def non_gaia(cls) -> List[InfoDatasetBase]:
        """
        Returns:
            A list of all the units excluding gaia only units (eg. militia)
        """
        return cls._gaia_filter(gaia_only=False)

    @classmethod
    def _gaia_filter(cls, gaia_only: bool) -> List[InfoDatasetBase]:
        """
        Args:
            gaia_only: if set to true, lists all gaia only units. If set to `False`, lists all units except gaia only units

        Returns:
            A list of either all gaia only units or a list of all units excluding gaia only units
        """
        result = []
        for x in cls:
            if x.IS_GAIA_ONLY == gaia_only:
                result.append(x)
        return result



from AoE2ScenarioParser.datasets.players import ColorId
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_group import RetrieverObjectLinkGroup


class PlayerResources(AoE2Object):
    """Object for handling a tile in the map."""

    _link_list = [
        RetrieverObjectLinkGroup("PlayerDataTwo", "resources[__index__]", group=[
            RetrieverObjectLink("food"),
            RetrieverObjectLink("wood"),
            RetrieverObjectLink("gold"),
            RetrieverObjectLink("stone"),
            RetrieverObjectLink("color", link="player_color"),
        ])
    ]

    def __init__(self, food: int, wood: int, gold: int, stone: int, color: int, **kwargs):
        super().__init__(**kwargs)

        self.food: int = food
        self.wood: int = wood
        self.gold: int = gold
        self.stone: int = stone
        self.color: ColorId = ColorId(color)



from __future__ import annotations

from typing import List


class TriggerCELock:
    """Object used to identify which conditions and effects should be locked from change"""
    def __init__(
            self,
            lock_conditions: bool = False,
            lock_effects: bool = False,
            lock_condition_type: List[int] | None = None,
            lock_effect_type: List[int] | None = None,
            lock_condition_ids: List[int] | None = None,
            lock_effect_ids: List[int] | None = None
    ):
        """
        Args:
            lock_conditions: Lock all conditions
            lock_effects: Lock all effects
            lock_condition_type: Lock certain condition types. Example: `ConditionId.OWN_OBJECTS`
            lock_effect_type: Lock certain effect types. Example: `EffectId.CREATE_OBJECT`
            lock_condition_ids: Lock certain conditions by their id
            lock_effect_ids: Lock certain effects by their id
        """
        if lock_condition_type is None:
            lock_condition_type = []
        if lock_effect_type is None:
            lock_effect_type = []
        if lock_condition_ids is None:
            lock_condition_ids = []
        if lock_effect_ids is None:
            lock_effect_ids = []

        self.lock_conditions = lock_conditions
        self.lock_effects = lock_effects
        self.lock_condition_type = lock_condition_type
        self.lock_effect_type = lock_effect_type
        self.lock_condition_ids = lock_condition_ids
        self.lock_effect_ids = lock_effect_ids



from __future__ import annotations

from AoE2ScenarioParser.exceptions.asp_exceptions import EndOfFileError


class IncrementalGenerator:
    """Similar to an actual generator and is used to return the bytes of a file sequentially"""
    def __init__(self, name: str, file_content: bytes, progress: int = 0):
        """
        Args:
            name: The full file path of the file to create the generator for
            file_content: The bytes of the specified file
            progress: Keeps track of how many bytes have been read from the generator.
        """
        self.name = name
        self.file_content: bytes = file_content
        self.progress = progress

    @classmethod
    def from_file(cls, filepath: str) -> IncrementalGenerator:
        """
        Creates and returns an instance of the IncrementalGenerator class from the given file

        Args:
            filepath: The path to the file to create the object from

        Returns:
            An instance of the IncrementalGenerator class
        """
        with open(filepath, 'rb') as f:
            file_content = f.read()
        return cls(filepath, file_content)

    def get_bytes(self, n: int, update_progress: bool = True) -> bytes:
        """
        Get the specified amount of next bytes

        Args:
            n: The number of bytes to get
            update_progress: If set to False, the counter for where to read the next set of bytes from won't be
                increased

        Returns:
            The specified number of bytes

        Raises:
            EndOfFileError: When the number of bytes requested is greater than the number of remaining bytes in the
                generator
        """
        if n <= 0:
            return b''
        result = self.file_content[self.progress:self.progress + n]
        if not result:
            remaining = len(self.get_remaining_bytes())
            raise EndOfFileError(f"End of file reached. (Requested: {n} bytes, only {remaining} left.")
        if update_progress:
            self.progress += n
        return result

    def get_remaining_bytes(self) -> bytes:
        """
        Get all of the remaining bytes in the generator

        Returns:
            The remaining bytes of the generator
        """
        result = self.file_content[self.progress:]
        self.progress = len(self.file_content) - 1
        return result

    def __repr__(self):
        return f"[IncrementalGenerator] Name: {self.name}\n\tProgress: {self.progress}/{len(self.file_content)}"



from __future__ import annotations

from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_group import RetrieverObjectLinkGroup


class Variable(AoE2Object):
    """Object for handling a variable."""

    _link_list = [
        RetrieverObjectLinkGroup("Triggers", "variable_data[__index__]", group=[
            RetrieverObjectLink("variable_id"),
            RetrieverObjectLink("name", link="variable_name"),
        ])
    ]

    def __init__(self, variable_id: int, name: str, **kwargs):
        self.variable_id = variable_id
        self.name = name

        super().__init__(**kwargs)



from typing import List

from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.data_objects.unit import Unit
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_group import RetrieverObjectLinkGroup


class PlayerUnits(AoE2Object):
    """Object for handling a tile in the map."""

    _link_list = [
        RetrieverObjectLinkGroup("Units", "players_units[__index__]", group=[
            RetrieverObjectLink("unit_count"),
            RetrieverObjectLink("units", process_as_object=Unit),
        ])
    ]

    def __init__(self, unit_count: int, units: List[Unit], **kwargs):
        super().__init__(**kwargs)

        self.unit_count: int = unit_count
        self.units: List[Unit] = units



from typing import List, Any

import AoE2ScenarioParser.datasets.conditions as condition_dataset
import AoE2ScenarioParser.datasets.effects as effect_dataset
from AoE2ScenarioParser.datasets.conditions import ConditionId
from AoE2ScenarioParser.datasets.effects import EffectId
from AoE2ScenarioParser.exceptions.asp_exceptions import UnsupportedAttributeError
from AoE2ScenarioParser.helper.helper import mutually_exclusive
from AoE2ScenarioParser.helper.list_functions import list_changed, update_order_array, hash_list
from AoE2ScenarioParser.helper.string_manipulations import add_tabs
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.data_objects.condition import Condition
from AoE2ScenarioParser.objects.data_objects.effect import Effect
from AoE2ScenarioParser.objects.support.new_condition import NewConditionSupport
from AoE2ScenarioParser.objects.support.new_effect import NewEffectSupport
from AoE2ScenarioParser.objects.support.trigger_object import TriggerComponent
from AoE2ScenarioParser.objects.support.uuid_list import UuidList
from AoE2ScenarioParser.scenarios.scenario_store import getters
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_group import RetrieverObjectLinkGroup


class Trigger(AoE2Object, TriggerComponent):
    """Object for handling a trigger."""

    _link_list = [
        RetrieverObjectLinkGroup("Triggers", "trigger_data[__index__]", group=[
            RetrieverObjectLink("name", link="trigger_name"),
            RetrieverObjectLink("description", link="trigger_description"),
            RetrieverObjectLink("description_stid", link="description_string_table_id"),
            RetrieverObjectLink("display_as_objective"),
            RetrieverObjectLink("short_description"),
            RetrieverObjectLink("short_description_stid", link="short_description_string_table_id"),
            RetrieverObjectLink("display_on_screen"),
            RetrieverObjectLink("description_order", link="objective_description_order"),
            RetrieverObjectLink("enabled"),
            RetrieverObjectLink("looping"),
            RetrieverObjectLink("header", link="make_header"),
            RetrieverObjectLink("mute_objectives"),
            RetrieverObjectLink("conditions", link="condition_data", process_as_object=Condition),
            RetrieverObjectLink("condition_order", link="condition_display_order_array"),
            RetrieverObjectLink("effects", link="effect_data", process_as_object=Effect),
            RetrieverObjectLink("effect_order", link="effect_display_order_array"),
        ]),
        RetrieverObjectLink("trigger_id", retrieve_history_number=0),
    ]

    def __init__(
            self,
            name: str,
            description: str = "",
            description_stid: int = -1,
            display_as_objective: int = 0,
            short_description: str = "",
            short_description_stid: int = -1,
            display_on_screen: int = 0,
            description_order: int = 0,
            enabled: int = 1,
            looping: int = 0,
            header: int = 0,
            mute_objectives: int = 0,
            conditions: List[Condition] = None,
            condition_order: List[int] = None,
            effects: List[Effect] = None,
            effect_order: List[int] = None,
            trigger_id: int = -1,
            **kwargs
    ):
        super().__init__(**kwargs)

        if conditions is None:
            conditions = []
        if condition_order is None:
            condition_order = []
        if effects is None:
            effects = []
        if effect_order is None:
            effect_order = []

        self.name: str = name
        self.description: str = description
        self.description_stid: int = description_stid
        self.display_as_objective: int = display_as_objective
        self.short_description: str = short_description
        self.short_description_stid: int = short_description_stid
        self.display_on_screen: int = display_on_screen
        self.description_order: int = description_order
        self.enabled: int = enabled
        self.looping: int = looping
        self.header: int = header
        self.mute_objectives: int = mute_objectives
        self._condition_hash = hash_list(conditions)
        self.conditions: List[Condition] = conditions
        self.condition_order: List[int] = condition_order
        self._effect_hash = hash_list(effects)
        self.effects: List[Effect] = effects
        self.effect_order: List[int] = effect_order
        self.trigger_id: int = trigger_id

        self.new_effect: NewEffectSupport = NewEffectSupport(self)
        self.new_condition: NewConditionSupport = NewConditionSupport(self)

        self._assign_new_ce_support()

    def _deepcopy_entry(self, k, v) -> Any:
        if k in ['new_effect', 'new_condition']:
            return None
        else:
            return super()._deepcopy_entry(k, v)

    def _assign_new_ce_support(self):
        """Assigns new `new_effect` and `new_condition` objects to this trigger"""
        self.new_effect = NewEffectSupport(self)
        self.new_condition = NewConditionSupport(self)

    @property
    def condition_order(self) -> List[int]:
        """The condition display order """
        if list_changed(self.conditions, self._condition_hash):
            update_order_array(self._condition_order, len(self.conditions))
            self._condition_hash = hash_list(self.conditions)
        return self._condition_order

    @condition_order.setter
    def condition_order(self, val):
        self._condition_order = val

    @property
    def effect_order(self) -> List[int]:
        """The effect display order """
        if list_changed(self.effects, self._effect_hash):
            update_order_array(self._effect_order, len(self.effects))
            self._effect_hash = hash_list(self.effects)
        return self._effect_order

    @effect_order.setter
    def effect_order(self, val):
        self._effect_order = val

    @property
    def conditions(self) -> List[Condition]:
        """All conditions in this trigger"""
        return self._conditions

    @conditions.setter
    def conditions(self, val: List[Condition]) -> None:
        self._conditions = UuidList(self._uuid, val)
        self.condition_order = list(range(0, len(val)))

    @property
    def effects(self) -> List[Effect]:
        """All effects in this trigger"""
        return self._effects

    @effects.setter
    def effects(self, val: List[Effect]) -> None:
        self._effects = UuidList(self._uuid, val)
        self.effect_order = list(range(0, len(val)))

    def _add_effect(self, effect_type: EffectId, ai_script_goal=None, armour_attack_quantity=None,
            armour_attack_class=None, quantity=None, tribute_list=None, diplomacy=None,
            object_list_unit_id=None, source_player=None, target_player=None, technology=None, string_id=None,
            display_time=None, trigger_id=None, location_x=None, location_y=None,
            location_object_reference=None, area_x1=None, area_y1=None, area_x2=None, area_y2=None,
            object_group=None, object_type=None, instruction_panel_position=None, attack_stance=None,
            time_unit=None, enabled=None, food=None, wood=None, stone=None, gold=None, item_id=None,
            flash_object=None, force_research_technology=None, visibility_state=None, scroll=None,
            operation=None, object_list_unit_id_2=None, button_location=None, ai_signal_value=None,
            object_attributes=None, variable=None, timer=None, facet=None, play_sound=None, message=None,
            player_color=None, sound_name=None, selected_object_ids=None, color_mood=None, reset_timer=None,
            object_state=None, action_type=None, ) -> Effect:
        """Used to add new effect to trigger. Please use trigger.new_effect.<effect_name> instead"""

        def get_default_effect_attributes(eff_type):
            """Gets the default effect attributes based on a certain effect type, with exception handling"""
            sv = getters.get_scenario_version(self._uuid)
            try:
                return effect_dataset.default_attributes[eff_type]
            except KeyError:
                effect = EffectId(eff_type)
                raise UnsupportedAttributeError(
                    f"The effect {effect.name} is not supported in scenario version {sv}"
                ) from None

        effect_defaults = get_default_effect_attributes(effect_type)
        effect_attr = {}
        for key, value in effect_defaults.items():
            effect_attr[key] = (locals()[key] if locals()[key] is not None else value)
        new_effect = Effect(**effect_attr, uuid=self._uuid)
        self.effects.append(new_effect)
        return new_effect

    def _add_condition(self, condition_type: ConditionId, quantity=None,
            attribute=None, unit_object=None, next_object=None, object_list=None,
            source_player=None, technology=None, timer=None, area_x1=None, area_y1=None, area_x2=None,
            area_y2=None, object_group=None, object_type=None, ai_signal=None, inverted=None, variable=None,
            comparison=None, target_player=None, unit_ai_action=None, xs_function=None, object_state=None,
            timer_id=None, victory_timer_type=None, include_changeable_weapon_objects=None,
    ) -> Condition:
        """Used to add new condition to trigger. Please use trigger.new_condition.<condition_name> instead"""

        def get_default_condition_attributes(cond_type):
            """Gets the default condition attributes based on a certain condition type, with exception handling"""
            sv = getters.get_scenario_version(self._uuid)
            try:
                return condition_dataset.default_attributes[cond_type]
            except KeyError:
                condition = ConditionId(cond_type)
                raise UnsupportedAttributeError(
                    f"The condition {condition.name} is not supported in scenario version {sv}"
                ) from None

        condition_defaults = get_default_condition_attributes(condition_type)
        condition_attr = {}
        for key, value in condition_defaults.items():
            condition_attr[key] = (locals()[key] if locals()[key] is not None else value)
        new_condition = Condition(**condition_attr, uuid=self._uuid)
        self.conditions.append(new_condition)
        return new_condition

    def get_effect(self, effect_index: int = None, display_index: int = None) -> Effect:
        """
        Retrieve an effect based on its (display) index

        Args:
            effect_index: The effect index of the effect to retrieve
            display_index: The display index of the effect to retrieve

        Raises:
            ValueError: If more than one parameter is used simultaneously
            IndexError: If the effect could not be found

        Returns:
            The wanted effect
        """
        if not mutually_exclusive(effect_index is not None, display_index is not None):
            raise ValueError(f"Please identify an effect using either effect_index or display_index.")

        if effect_index is None:
            effect_index = self.effect_order[display_index]

        return self.effects[effect_index]

    def get_condition(self, condition_index: int = None, display_index: int = None) -> Condition:
        """
        Retrieve a condition based on its (display) index

        Args:
            condition_index: The condition index of the condition to retrieve
            display_index: The display index of the condition to retrieve

        Raises:
            ValueError: If more than one parameter is used simultaneously
            IndexError: If the condition could not be found

        Returns:
            The wanted condition
        """
        if not mutually_exclusive(condition_index is not None, display_index is not None):
            raise ValueError(f"Please identify a condition using either condition_index or display_index.")

        if condition_index is None:
            condition_index = self.condition_order[display_index]

        return self.conditions[condition_index]

    def remove_effect(self, effect_index: int = None, display_index: int = None, effect: Effect = None) -> None:
        """
        Remove an effect from this trigger

        Args:
            effect_index: The effect index of the condition to remove
            display_index: The display index of the condition to remove
            effect: A reference to the effect object to remove from this trigger

        Raises:
            ValueError: If more than one parameter is used simultaneously
        """
        if not mutually_exclusive(effect_index is not None, display_index is not None, effect is not None):
            raise ValueError(f"Please identify an effect using either effect_index, display_index or effect.")

        if effect is not None:
            effect_index = self.effects.index(effect)
        if effect_index is None:
            effect_index = self.effect_order[display_index]

        del self.effects[effect_index]

    def remove_condition(self, condition_index: int = None, display_index: int = None, condition: Condition = None) \
            -> None:
        """
        Remove a condition from this trigger

        Args:
            condition_index: The condition index of the condition to remove
            display_index: The display index of the condition to remove
            condition: A reference to the condition object to remove from this trigger

        Raises:
            ValueError: If more than one parameter is used simultaneously
        """
        if not mutually_exclusive(condition_index is not None, display_index is not None, condition is not None):
            raise ValueError(f"Please identify a condition using either condition_index, display_index or condition.")

        if condition is not None:
            condition_index = self.conditions.index(condition)

        if condition_index is None:
            condition_index = self.condition_order[display_index]

        del self.conditions[condition_index]

    def get_content_as_string(self, include_trigger_definition: bool = False) -> str:
        """
        Create a human-readable string showcasing all content of this trigger.
        This includes all the content within the conditions and effects of this trigger.

        This is also the function that is called when doing: `print(trigger)`

        Args:
            include_trigger_definition: If the trigger meta-data should be added by this function

        Returns:
            The created string
        """
        return_string = ""

        data_tba = {
            'enabled': self.enabled != 0,
            'looping': self.looping != 0
        }

        if self.description != "":
            data_tba['description'] = f"'{self.description}'"
        if self.description_stid != -1:
            data_tba['description_stid'] = self.description_stid
        if self.short_description != "":
            data_tba['short_description'] = f"'{self.short_description}'"
        if self.short_description_stid != -1:
            data_tba['short_description_stid'] = self.short_description_stid
        if self.display_as_objective != 0:
            data_tba['display_as_objective'] = (self.display_as_objective != 0)
        if self.display_on_screen != 0:
            data_tba['display_on_screen'] = (self.display_on_screen != 0)
        if self.description_order != 0:
            data_tba['description_order'] = self.description_order
        if self.header != 0:
            data_tba['header'] = (self.header != 0)
        if self.mute_objectives != 0:
            data_tba['mute_objectives'] = (self.mute_objectives != 0)

        for key, value in data_tba.items():
            return_string += f"{key}: {value}\n"

        if len(self.condition_order) > 0:
            return_string += "conditions:\n"
            for c_display_order, condition_id in enumerate(self.condition_order):
                condition = self.conditions[condition_id]

                name = f"Unknown ({condition.condition_type})"
                if condition.condition_type in condition_dataset.condition_names:
                    name = condition_dataset.condition_names[condition.condition_type]

                return_string += f"\t{name} [Index: {condition_id}, Display: {c_display_order}]:\n"
                return_string += add_tabs(condition.get_content_as_string(), 2)

        if len(self.effect_order) > 0:
            return_string += "effects:\n"
            for e_display_order, effect_id in enumerate(self.effect_order):
                effect = self.effects[effect_id]

                name = f"Unknown ({effect.effect_type})"
                if effect.effect_type in effect_dataset.effect_names:
                    name = effect_dataset.effect_names[effect.effect_type]

                return_string += f"\t{name} [Index: {effect_id}, Display: {e_display_order}]:\n"
                return_string += add_tabs(effect.get_content_as_string(), 2)

        if include_trigger_definition:
            return f"\"{self.name}\" [Index: {self.trigger_id}]\n" + add_tabs(return_string, 1)
        return return_string

    def __str__(self) -> str:
        return f"[Trigger] {self.get_content_as_string(include_trigger_definition=True)}"



from __future__ import annotations


class DataType:
    """
    This class is used to represent the data type of the value being retrieved. Every retriever has this as an attribute
    """
    __slots__ = [
        'var',
        '_repeat',
        'log_value',
        'type',
        'length',
        '_debug_retriever_name'
    ]

    _debug_retriever_name: str

    def __init__(self, var: str = "0", repeat: int = 1, log_value: bool = False, type_length: tuple = None):
        """
        Args:
            var: string representation of the datatype
            repeat: the amount of times a value is repeated
            log_value: to log the value of the datatype when repeat is set
            type_length: a tuple containing the type and length of the datatype

        Descriptions:
            var:
                Almost always a string representation of the data you want to retrieve.
                This project uses the types from: https://dderevjanik.github.io/agescx/formatscx/#format

                - **s**: Signed integer (parsed as little endian)
                - **u**: Unsigned integer (parsed as little endian)
                - **f**: float
                - **c**: Character string of fixed length
                - **str**: Character string of variable length. This type will read the number in bits given and parse it as
                  an int. The number retrieved from it will be the amount of bytes read as a character string.
                - **Struct:<StructName>**: This is like a class, and is a collection of related information. Structs may
                  contain sub-structs and other simpler data types. This will parse all DataType values in the Struct
                  subclass. This can be handy for when blocks of data are repeated.
                - **(Empty)**: Is interpreted as regular byte data. In this project the '' is converted to 'data'

                Each datatype is followed by a number that tells how many bits/bytes are used to store it.

                - The types **s, u, f & str** use an amount of bits to specify their size.
                - The types **c & data** use bytes to specify their size.

                For example:
                    * **s16**: A 16 bit signed integer.
                    * **f32**: A 32 bit floating point number.
                    * **c4**: A 32 bit (4 bytes) character string.
                    * **str16**: A 16 bit integer will be parsed (n). Now n bytes will be read as character string.
                    * **TerrainStruct**: The TerrainStruct will be instantiated and DataTypes from that struct will be
                      loaded in it's place.

            repeat:
                The amount of times the above datatype needs to be repeated

            log_value (Default: False):
                A boolean value indicating if the datatype should be logged

            type_length (Default: None):
                A tuple specifying the type and length of a retriever
        """
        self.log_value = log_value
        self.var = var
        self.repeat = repeat

        if type_length is not None:
            self.type, self.length = type_length
        else:
            self.type, self.length = datatype_to_type_length(self.var)

    def get_struct_name(self) -> str:
        """
        Returns the name of the datatype structure

        Returns:
            The name of the DataType structure

        Raises:
            ValueError: if the datatype is not a struct
        """
        if self.var.startswith("struct:"):
            return self.var[7:]
        raise ValueError("Datatype var is not a struct")

    @property
    def type_and_length(self) -> tuple:
        """A tuple of the type and length of the datatype"""
        return self.type, self.length

    @property
    def repeat(self) -> int:
        """The amount of times the datatype repeats"""
        return self._repeat

    @repeat.setter
    def repeat(self, value: int):
        """
        The setter for the repeat value of the datatype

        Args:
            value: The value to set repeat to. The value of an assignment is automatically passed as this argument
        """
        if type(value) is not int:
            raise TypeError(f"Repeat value must be an integer, now: {value}")

        if self.log_value and hasattr(self, '_debug_retriever_name'):
            print(f"[DataType] {self._debug_retriever_name} 'repeat' set to {value} (was: {self._repeat})")
        self._repeat = value

    def to_simple_string(self) -> str:
        """
        Get a string representation of the datatype containing its repeat value and the var type

        Returns:
            A string representation of the datatype
        """
        return f"{self._repeat} * {self.var}"

    def duplicate(self) -> DataType:
        """
        Create another datatype instance with the same values as this one

        Returns:
            A datatype instance with the same values as this one
        """
        return DataType(
            var=self.var,
            repeat=self.repeat,
            log_value=self.log_value,
            type_length=self.type_and_length
        )

    def __repr__(self):
        return f"[DataType]  " + self.to_simple_string()


def datatype_to_type_length(var: str) -> tuple:
    """
    Get the type and length of a datatype

    Args:
        var: The datatype string

    Returns:
        The type and length of a datatype. So: 'int32' returns 'int', 32.

    Raises:
        ValueError: if the variable type is not in [s, u, f, c, str, data]
    """
    if var[:7] == "struct:":
        return "struct", 0

    # Filter numbers out for length, filter text for type
    var_len = int(''.join(filter(str.isnumeric, var)))
    var_type = ''.join(filter(str.isalpha, var))

    if var_type == '':
        var_type = "data"

    if var_type not in datatype_types:
        raise ValueError(f"Unknown variable type '{var_type}'")

    # Divide by 8, and parse from float to int
    if var_type not in ["c", "data"]:
        var_len = int(var_len / 8)

    return var_type, var_len


datatype_types = [
    "s",  # Signed int
    "u",  # Unsigned int
    "f",  # FloatingPoint
    "c",  # Character string
    "str",  # Variable length string
    "data",  # Data (Can be changed by used using bytes_to_x functions)
]



from typing import List

from AoE2ScenarioParser.objects.data_objects.terrain_tile import TerrainTile
from AoE2ScenarioParser.objects.managers.map_manager import MapManager
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_group import RetrieverObjectLinkGroup
from AoE2ScenarioParser.sections.retrievers.support import Support


class MapManagerDE(MapManager):
    """Manager of all DE map related features"""
    _link_list = [
        RetrieverObjectLinkGroup("Map", group=[
            RetrieverObjectLink("map_color_mood"),
            RetrieverObjectLink("collide_and_correct"),
            RetrieverObjectLink("villager_force_drop", support=Support(since=1.37)),
            RetrieverObjectLink("map_width"),
            RetrieverObjectLink("map_height"),
            RetrieverObjectLink("terrain", link="terrain_data", process_as_object=TerrainTile),
        ])
    ]

    def __init__(self,
                 map_color_mood: str,
                 collide_and_correct: bool,
                 villager_force_drop: bool,
                 map_width: int,
                 map_height: int,
                 terrain: List[TerrainTile],
                 **kwargs,
                 ):
        super().__init__(map_width, map_height, terrain, **kwargs)

        self.map_color_mood: str = map_color_mood
        self.collide_and_correct: bool = collide_and_correct
        self.villager_force_drop: bool = villager_force_drop



from __future__ import annotations

import json
import time
import zlib
from pathlib import Path
from typing import Dict, TYPE_CHECKING, TypeVar, Type, Any
from uuid import uuid4, UUID

import AoE2ScenarioParser.datasets.conditions as conditions
import AoE2ScenarioParser.datasets.effects as effects
from AoE2ScenarioParser import settings
from AoE2ScenarioParser.datasets.game_variant import ScenarioVariant
from AoE2ScenarioParser.exceptions.asp_exceptions import InvalidScenarioStructureError, UnknownScenarioStructureError, \
    UnknownStructureError, UnsupportedVersionError
from AoE2ScenarioParser.exceptions.asp_warnings import IncorrectVariantWarning
from AoE2ScenarioParser.helper.bytes_conversions import bytes_to_int
from AoE2ScenarioParser.helper.incremental_generator import IncrementalGenerator
from AoE2ScenarioParser.helper.printers import s_print, color_string, warn
from AoE2ScenarioParser.helper.string_manipulations import create_textual_hex
from AoE2ScenarioParser.helper.version_check import python_version_check
from AoE2ScenarioParser.objects.aoe2_object_manager import AoE2ObjectManager
from AoE2ScenarioParser.objects.managers.map_manager import MapManager
from AoE2ScenarioParser.objects.managers.message_manager import MessageManager
from AoE2ScenarioParser.objects.managers.player_manager import PlayerManager
from AoE2ScenarioParser.objects.managers.trigger_manager import TriggerManager
from AoE2ScenarioParser.objects.managers.unit_manager import UnitManager
from AoE2ScenarioParser.scenarios.scenario_debug.compare import debug_compare
from AoE2ScenarioParser.scenarios.scenario_store import store
from AoE2ScenarioParser.scenarios.support.object_factory import ObjectFactory
from AoE2ScenarioParser.scenarios.support.scenario_actions import ScenarioActions
from AoE2ScenarioParser.sections.aoe2_file_section import AoE2FileSection

if TYPE_CHECKING:
    from AoE2ScenarioParser.objects.aoe2_object import AoE2Object

S: TypeVar = TypeVar('S', bound='AoE2Scenario')
"""
A type variable (generic) that represents an instance of the AoE2Scenario class or any of its 
subclasses (e.g. `AoE2DEScenario`) 
"""


class AoE2Scenario:
    """All scenario objects are derived from this class"""

    @property
    def trigger_manager(self) -> TriggerManager:
        """The trigger manager of the scenario"""
        return self._object_manager.managers['Trigger']

    @property
    def unit_manager(self) -> UnitManager:
        """The unit manager of the scenario"""
        return self._object_manager.managers['Unit']

    @property
    def map_manager(self) -> MapManager:
        """The map manager of the scenario"""
        return self._object_manager.managers['Map']

    @property
    def player_manager(self) -> PlayerManager:
        """The player manager of the scenario"""
        return self._object_manager.managers['Player']

    @property
    def message_manager(self) -> MessageManager:
        """The message manager of the scenario"""
        return self._object_manager.managers['Message']

    @property
    def scenario_version_tuple(self) -> tuple[int, ...]:
        return tuple(map(int, self.scenario_version.split('.')))

    def __init__(
        self,
        game_version: str,
        scenario_version: str,
        source_location: str,
        name: str,
        variant: ScenarioVariant | None = None
    ):
        # Scenario meta info
        self.game_version: str = game_version
        self.scenario_version: str = scenario_version
        self.source_location: str = source_location
        self._variant: ScenarioVariant | None = variant
        self._time_start: float = time.time()

        # Actual scenario content
        self.structure: Dict = {}
        self.sections: Dict[str, AoE2FileSection] = {}
        self._object_manager: AoE2ObjectManager | None = None

        # For Scenario Store functionality
        self.name: str = name
        self.uuid: UUID = uuid4()
        store.register_scenario(self)

        # Actions through the scenario
        self.new: ObjectFactory = ObjectFactory(self.uuid)
        self.actions: ScenarioActions = ScenarioActions(self.uuid)

        # Used in debug functions
        self._file = None
        self._file_header = None
        self._decompressed_file_data = None

    @classmethod
    def from_file(
        cls: Type[S],
        path: str,
        game_version: str,
        name: str = ""
    ) -> S:
        """
        Creates and returns an instance of the AoE2Scenario class from the given scenario file

        Args:
            path: The path to the scenario file to create the object from
            game_version: The version of the game to create the object for
            name: The name given to this scenario (defaults to the filename without extension)

        Returns:
            An instance of the AoE2Scenario class (or any of its subclasses) which is the object representation of
                the given scenario file
        """
        python_version_check()

        filepath = Path(path)
        if not filepath.is_file():
            raise ValueError(f"Unable to read file from path '{filepath}'")

        name = name or filepath.stem

        s_print(f"Reading file: " + color_string(f"'{path}'", "magenta"), final=True, time=True, newline=True)
        s_print("Reading scenario file...")
        igenerator = IncrementalGenerator.from_file(path)
        s_print("Reading scenario file finished successfully.", final=True, time=True)

        scenario_version = _get_file_version(igenerator)
        scenario_variant = _get_scenario_variant(igenerator)

        scenario: S = cls(game_version, scenario_version, source_location=path, name=name, variant=scenario_variant)

        variant: str = 'Unknown' if scenario.variant is None else scenario_variant.to_display_name()

        # Log game and scenario version
        s_print("\n############### Attributes ###############", final=True, color="blue")
        s_print(f">>> Game version: '{scenario.game_version}'", final=True, color="blue")
        s_print(f">>> Scenario version: {scenario.scenario_version}", final=True, color="blue")
        s_print(f">>> Scenario variant: '{variant}'", final=True, color= "blue")
        s_print("##########################################", final=True, color="blue")

        s_print(f"Loading scenario structure...", time=True, newline=True)
        scenario._load_structure()
        _initialise_version_dependencies(scenario.game_version, scenario.scenario_version)
        s_print(f"Loading scenario structure finished successfully.", final=True, time=True)

        # scenario._initialize(igenerator)
        s_print("Parsing scenario file...", final=True, time=True)
        scenario._load_header_section(igenerator)
        scenario._load_content_sections(igenerator)
        s_print(f"Parsing scenario file finished successfully.", final=True, time=True)
        scenario.igenerator = igenerator

        scenario._object_manager = AoE2ObjectManager(scenario.uuid)
        scenario._object_manager.setup()

        return scenario

    @staticmethod
    def get_scenario(
        uuid: UUID = None,
        obj: 'AoE2Object' = None,
        name: str = None
    ) -> S:
        """
        Get scenario through a UUID, a related object or the name of a scenario.

        Args:
            uuid: The UUID of the scenario
            obj: An object related to a scenario
            name: The name of a scenario

        Returns:
            The scenario based on the given identifier, or `None`
        """
        return store.get_scenario(uuid=uuid, obj=obj, name=name)

    def _load_structure(self) -> None:
        """
        Loads the structure json for the scenario and game version specified into self.structure

        Raises:
            ValueError: if the game or scenario versions are not set
        """
        if self.game_version == "???" or self.scenario_version == "???":
            raise ValueError("Both game and scenario version need to be set to load structure")
        self.structure = _get_structure(self.game_version, self.scenario_version)

    def _load_header_section(self, raw_file_igenerator: IncrementalGenerator) -> None:
        """
        Reads and adds the header file section to the sections dict of the scenario.

        The header is stored decompressed and is the first thing in the scenario file. It is meta data for the scenario
        file and thus needs to be read before everything else (also the reason why its stored decompressed).

        Args:
            raw_file_igenerator: The generator to read the header section from
        """
        header = self._create_and_load_section('FileHeader', raw_file_igenerator)
        self._file_header = raw_file_igenerator.file_content[:raw_file_igenerator.progress]
        self._add_to_sections(header)

    def _load_content_sections(self, raw_file_igenerator: IncrementalGenerator) -> None:
        """
        Reads and adds all the remaining file sections from the structure file to the sections dict of the
        scenario.

        The sections after the header are compressed and are first decompressed using the -zlib.MAX_WBITS algorithm.

        Args:
            raw_file_igenerator: The generator to read the file sections from
        """
        self._decompressed_file_data = _decompress_bytes(raw_file_igenerator.get_remaining_bytes())

        data_igenerator = IncrementalGenerator(name='Scenario Data', file_content=self._decompressed_file_data)

        for section_name in self.structure.keys():
            if section_name == "FileHeader":
                continue
            try:
                section = self._create_and_load_section(section_name, data_igenerator)
                self._add_to_sections(section)
            except (ValueError, TypeError) as e:
                print(f"\n[{e.__class__.__name__}] AoE2Scenario.parse_file: \n\tSection: {section_name}\n")
                self.write_error_file(trail_generator=data_igenerator)
                raise e

    def _create_and_load_section(self, name: str, igenerator: IncrementalGenerator) -> AoE2FileSection:
        """
        Initialises a file section from its name and fills its retrievers with data from the given generator

        Args:
            name: The name of the file section
            igenerator: The generator to fill the data from

        Returns:
            An AoE2FileSection representing the given section name with its data initialised from the generator
        """
        s_print(f"\t🔄 Parsing {name}...", color="yellow")
        section = AoE2FileSection.from_structure(name, self.structure.get(name), self.uuid)
        s_print(f"\t🔄 Gathering {name} data...", color="yellow")
        section.set_data_from_generator(igenerator)
        s_print(f"\t✔ {name}", final=True, color="green")
        return section

    def _add_to_sections(self, section: AoE2FileSection) -> None:
        """
        Adds the given section to the sections dictionary

        Args:
            section: The section to add to the sections dictionary
        """
        self.sections[section.name] = section

    def remove_store_reference(self) -> None:
        """
        This function is **DEPRECATED**. No replacement is necessary as the store now uses weak references.
        You can safely remove the call to this function.

        --- Legacy docstring ---

        Removes the reference to this scenario object from the scenario store. Useful (~a must) when reading many
        scenarios in a row without needing earlier ones. Python likes to take up a lot of memory.
        Removing all references to an object will cause the memory to be cleared up.

        Warning: Remove all other references too!
            When using this function it's important to remove all other references to the scenario.
            So if save it in a dict or list, remove it from it.
            If you have variables referencing this scenario that you won't need anymore (and won't overwrite) delete
            them using: `del varname`.
        """
        warn("This function is DEPRECATED as the store now uses weak references. \n"
             "You can safely remove the call to this function.", DeprecationWarning)
        store.remove_scenario(self.uuid)

    def commit(self) -> None:
        """Commit the changes to the retriever backend made within the managers."""
        self._object_manager.reconstruct()

    """ ##########################################################################################
    ####################################### Write functions ######################################
    ########################################################################################## """

    def write_to_file(self, filename: str, skip_reconstruction: bool = False) -> None:
        """
        Writes the scenario to a new file with the given filename

        Args:
            filename: The location to write the file to
            skip_reconstruction: If reconstruction should be skipped. If true, this will ignore all changes made
                using the managers (For example all changes made using trigger_manager).

        Raises:
            ValueError: if the setting DISABLE_ERROR_ON_OVERWRITING_SOURCE is not disabled and the source filename is
                the same as the filename being written to
        """
        self._write_from_structure(filename, skip_reconstruction)

    def _write_from_structure(self, filename: str, skip_reconstruction: bool = False) -> None:
        """
        Writes the scenario to a new file with the given filename

        Args:
            filename: The location to write the file to
            skip_reconstruction: If reconstruction should be skipped. If true, this will ignore all changes made
                using the managers (For example all changes made using trigger_manager).

        Raises:
            ValueError: if the setting DISABLE_ERROR_ON_OVERWRITING_SOURCE is not disabled and the source filename is
                the same as the filename being written to
        """
        if settings.ALLOW_OVERWRITING_SOURCE and self.source_location == filename:
            raise ValueError("Overwriting the source scenario file is discouraged & disallowed. ")
        if not skip_reconstruction:
            self.commit()

        self._validate_scenario_variant()

        s_print("File writing from structure started...", final=True, time=True, newline=True)
        binary = _get_file_section_data(self.sections.get('FileHeader'))

        binary_list_to_be_compressed = []
        for file_part in self.sections.values():
            if file_part.name == "FileHeader":
                continue
            binary_list_to_be_compressed.append(_get_file_section_data(file_part))

        compressed = _compress_bytes(b''.join(binary_list_to_be_compressed))

        with open(filename, 'wb') as f:
            f.write(binary + compressed)

        etime = round(time.time() - self._time_start, 2)
        s_print("File writing finished successfully.", final=True, time=True)
        s_print(f"File successfully written to: " + color_string(f"'{filename}'", "magenta"), final=True, time=True)
        s_print(f"Execution time from scenario read: {etime}s", final=True, time=True)

    def write_error_file(self, filename: str = "error_file.txt", trail_generator: IncrementalGenerator = None) -> None:
        """
        Outputs the contents of the entire scenario file in a readable format. An example of the format is given below::

            ########################### units (1954 * struct:UnitStruct)
            ############ UnitStruct ############  [STRUCT]
            00 00 70 42                 x (1 * f32): 60.0
            00 00 70 42                 y (1 * f32): 60.0
            00 00 00 00                 z (1 * f32): 0.0
            52 05 00 00                 reference_id (1 * s32): 1362
            89 02                       unit_const (1 * u16): 649
            02                          status (1 * u8): 2
            00 00 00 00                 rotation (1 * f32): 0.0
            00 00                       initial_animation_frame (1 * u16): 0
            ff ff ff ff                 garrisoned_in_id (1 * s32): -1

        Args:
            filename: The filename to write the error file to
            trail_generator: Write all the bytes remaining in this generator as a trail
        """
        self._debug_byte_structure_to_file(filename=filename, trail_generator=trail_generator)

    """ #############################################
    ############### Variant functions ###############
    ############################################# """

    @property
    def variant(self) -> ScenarioVariant:
        return self._variant

    @variant.setter
    def variant(self, value: None | str | int | ScenarioVariant):
        if value is None:
            self._variant = None
            return
        elif isinstance(value, ScenarioVariant):
            self._variant = value
        elif isinstance(value, int):
            self._variant = ScenarioVariant(value)
        elif isinstance(value, str):
            self._variant = ScenarioVariant[value.upper()]
        else:
            raise ValueError(f"Incorrect value used for setting scenario variant: '{value}'")

        self._update_variant_retrievers()
        self._validate_scenario_variant()

    def _validate_scenario_variant(self):
        if self.variant is None:
            self._warn_variant_unknown()
            return

        # If the header has been adjusted manually (solution before this functionality went live)
        if self.sections["FileHeader"].unknown_value_2 != self.variant.value:
            self.variant = self.sections["FileHeader"].unknown_value_2

        if self.variant == ScenarioVariant.ROR and self.scenario_version_tuple < (1, 49):
            raise UnsupportedVersionError(
                f"\n\nScenarios with a version below 1.49 (currently: {self.scenario_version}) cannot be written as "
                f"Return of Rome scenarios.\n"
                "Upgrade the scenario by saving it in the in-game editor before converting it."
            )

        if self.variant not in [ScenarioVariant.AOE2, ScenarioVariant.ROR] and settings.SHOW_VARIANT_WARNINGS:
            applicants = self.variant.applicants()
            name = self.variant.name

            warn(
                f"Having the scenario variant set to '{name}' (applies to: '{applicants}') will cause the scenario "
                f"to not be visible within the Definitive Edition.\n"
                f"If this is unintentional, you can set `scenario.variant` to 'aoe2' or 'ror'`\n"
                f"If this is intentional, you can disable this warning using the "
                f"setting: `SHOW_VARIANT_WARNINGS`", IncorrectVariantWarning
            )

    def _warn_variant_unknown(self):
        if not settings.SHOW_VARIANT_WARNINGS:
            return

        warn(
            f"The current scenario variant is unknown. It's possible this will cause the scenario to be hidden.\n"
            f"If this is unintentional, you can set `scenario.variant` to 'aoe2' or 'ror' to make sure it's visible.\n"
            f"If this is intentional, you can disable this warning using the "
            f"setting: `SHOW_VARIANT_WARNINGS`", IncorrectVariantWarning
        )

    def _update_variant_retrievers(self):
        if self.variant.value == self.sections["FileHeader"].unknown_value_2:
            return

        dlcs = {
            ScenarioVariant.LEGACY: [],
            ScenarioVariant.AOE2: [2, 3, 4, 5, 6, 7, 8, 9, 10, 12],
            ScenarioVariant.ROR: [11]
        }[self.variant]

        self.sections["FileHeader"].unknown_value_2 = self.variant.value
        self.sections["FileHeader"].amount_of_unknown_numbers = len(dlcs)
        self.sections["FileHeader"].unknown_numbers = dlcs

    """ #############################################
    ################ Debug functions ################
    ############################################# """

    def _debug_compare(
            self,
            other: AoE2Scenario,
            filename: str = "differences.txt",
            commit: bool = False,
            *,
            allow_multiple_versions: bool = False
    ) -> None:
        """
        Compare a scenario to a given scenario and report the differences found

        Args:
            other: The scenario to compare it to
            filename: The debug file to write the differences to (Defaults to "differences.txt")
            commit: If the scenarios need to commit their manager changes before comparing (Defaults to False)
            allow_multiple_versions: Allow comparison between multiple versions. (Please note that this is not tested
                thoroughly at all)
        """
        debug_compare(self, other, filename, commit, allow_multiple_versions=allow_multiple_versions)

    def _debug_write_from_source(self, filename: str, datatype: str, write_bytes: bool = True) -> None:
        """
        Writes the decompressed scenario file as bytes or as hex text

        Args:
            filename: The filename to write to
            datatype: these are flags that indicate which parts of the file to include in the output. 'd' for
                decompressed file data, 'f' for the file, and 'h' for the header. Note: Only 'd' actually works at this
                time
            write_bytes: boolean to determine if the file needs to be written as bytes or hex text form
        """
        s_print("File writing from source started with attributes " + datatype + "...")
        file = open(filename, "wb" if write_bytes else "w")
        selected_parts = []
        for t in datatype:
            if t == "f":
                selected_parts.append(self._file)
            elif t == "h":
                selected_parts.append(self._file_header)
            elif t == "d":
                selected_parts.append(self._decompressed_file_data)
        parts = None
        for part in selected_parts:
            if parts is None:
                parts = part
                continue
            parts += part
        file.write(parts if write_bytes else create_textual_hex(parts.hex()))
        file.close()
        s_print("File writing finished successfully.")

    def _debug_byte_structure_to_file(self, filename, trail_generator: IncrementalGenerator = None, commit=False):
        """
        Outputs the contents of the entire scenario file in a readable format. An example of the format is given below::

            ########################### units (1954 * struct:UnitStruct)
            ############ UnitStruct ############  [STRUCT]
            00 00 70 42                 x (1 * f32): 60.0
            00 00 70 42                 y (1 * f32): 60.0
            00 00 00 00                 z (1 * f32): 0.0
            52 05 00 00                 reference_id (1 * s32): 1362
            89 02                       unit_const (1 * u16): 649
            02                          status (1 * u8): 2
            00 00 00 00                 rotation (1 * f32): 0.0
            00 00                       initial_animation_frame (1 * u16): 0
            ff ff ff ff                 garrisoned_in_id (1 * s32): -1

        Args:
            filename: The filename to write the error file to
            trail_generator: Write all the bytes remaining in this generator as a trail
            commit: If the managers should commit their changes before writing this file.
        """
        if commit and hasattr(self, '_object_manager'):
            self.commit()

        s_print("Writing structure to file...", final=True, time=True, newline=True)

        result = []
        for section in self.sections.values():
            s_print(f"\t🔄 Writing {section.name}...", color="yellow")
            result.append(section.get_byte_structure_as_string())
            s_print(f"\t✔ {section.name}", final=True, color="green")

        if trail_generator is not None:
            s_print("\tWriting trail...")
            trail = trail_generator.get_remaining_bytes()

            result.append(f"\n\n{'#' * 27} TRAIL ({len(trail)})\n\n")
            result.append(create_textual_hex(trail.hex(), space_distance=2, enter_distance=24))
            s_print("\tWriting trail finished successfully.", final=True)

        with open(filename, 'w', encoding=settings.MAIN_CHARSET) as f:
            f.write(''.join(result))
        s_print("Writing structure to file finished successfully.", final=True, time=True)


def _initialise_version_dependencies(game_version: str, scenario_version: str) -> None:
    """
    Initialises the data for the condition and effect objects (IDs, defaults, etc.) for the given scenario
    and game versions

    Args:
        game_version: The version of the game to initialise the dependencies for
        scenario_version: The version of the scenario to initialise the dependencies for
    """
    condition_json = _get_version_dependant_structure_file(game_version, scenario_version, "conditions")

    for condition_id, structure in condition_json.items():
        condition_id = int(condition_id)

        if condition_id == -1:
            conditions.attribute_presentation[condition_id] = structure['attribute_presentation']
            continue

        conditions.condition_names[condition_id] = structure['name']
        conditions.default_attributes[condition_id] = structure['default_attributes']
        conditions.attributes[condition_id] = structure['attributes']
        conditions.attribute_presentation[condition_id] = structure.get('attribute_presentation', {})

    effect_json = _get_version_dependant_structure_file(game_version, scenario_version, "effects")

    for effect_id, structure in effect_json.items():
        effect_id = int(effect_id)

        if effect_id == -1:
            effects.attribute_presentation[effect_id] = structure['attribute_presentation']
            continue

        effects.effect_names[effect_id] = structure['name']
        effects.default_attributes[effect_id] = structure['default_attributes']
        effects.attributes[effect_id] = structure['attributes']
        effects.attribute_presentation[effect_id] = structure.get('attribute_presentation', {})


def _get_file_section_data(file_section: AoE2FileSection) -> bytes:
    """
    Converts the data of the given file section into bytes, and also prints that the conversion is happening to the
    console.

    Args:
        file_section: The file section to convert to bytes

    Returns:
        Bytes for all the data in the given file section
    """
    s_print(f"\t🔄 Reconstructing {file_section.name}...", color="yellow")
    value = file_section.get_data_as_bytes()
    s_print(f"\t✔ {file_section.name}", final=True, color="green")
    return value


def _get_file_version(generator: IncrementalGenerator):
    """
    Get first 4 bytes of a file, which contains the version of the scenario
    
    Args:
        generator: An IncrementalGenerator object of a scenario file

    Returns:
        A string which is the version of the scenario file
    """
    return generator.get_bytes(4, update_progress=False).decode('ASCII')


def _get_scenario_variant(generator: IncrementalGenerator) -> ScenarioVariant | None:
    """
    Find the bytes of the retriever before the scenario variant bytes. Then use that reference to find out which variant
    it is. If it's unable to find it, None is returned.

    Args:
        generator: An IncrementalGenerator object of a scenario file

    Returns:
        A ScenarioVariant object or None if it couldn't find the variant
    """
    try:
        index = generator.file_content.index(b"\xE8\x03\x00\x00")
    except ValueError:
        return None  # Unable to find sequence

    value = generator.file_content[index + 4:index + 8]
    value = bytes_to_int(value, signed = False)
    try:
        return ScenarioVariant(value)
    except ValueError:
        return None  # Variant not in Enum


def _decompress_bytes(file_content: bytes) -> bytes:
    """
    Decompress the given bytes using the -zlib.MAX_WBITS algorithm

    Args:
        file_content: The bytes to decompress

    Returns:
        The decompressed bytes
    """
    return zlib.decompress(file_content, -zlib.MAX_WBITS)


def _compress_bytes(file_content: bytes) -> bytes:
    """
    Compress the given bytes using the -zlib.MAX_WBITS algorithm. View this link for additional information:
    https://stackoverflow.com/questions/3122145/zlib-error-error-3-while-decompressing-incorrect-header-check/22310760#22310760

    Args:
        file_content: The bytes to compress

    Returns:
        The compressed bytes
    """
    deflate_obj = zlib.compressobj(9, zlib.DEFLATED, -zlib.MAX_WBITS)
    compressed = deflate_obj.compress(file_content) + deflate_obj.flush()
    return compressed


def _get_version_directory_path() -> Path:
    """
    Get the path of the AoE2ScenarioParser/versions directory. Simply '../../versions' is not used as that is not OS
    friendly.

    Returns:
        The full path of the AoE2ScenarioParser/versions directory.
    """
    return Path(__file__).parent.parent / 'versions'


def _get_version_dependant_structure_file(game_version: str, scenario_version: str, name: str) -> dict:
    """
    Returns a structure file dependant on the version of the scenario (AND game version).
    Files are retrieved based on the game and scenario version given.

    Args:
        game_version: The version of the game to return the structure file for
        scenario_version: The scenario version to return the structure file for
        name: The name of the structure file being requested

    Returns:
        a dict representation of the structure file requested

    Raises:
        UnknownStructureError: if a json associated with the name and versions specified is not found
    """
    try:
        vdir = _get_version_directory_path()
        with (vdir / game_version / f'v{scenario_version}' / f'{name}.json').open() as structure_file:
            return json.load(structure_file)
    except FileNotFoundError:  # Unsupported version
        v = f"{game_version}:{scenario_version}"
        raise UnknownStructureError(f"The structure {name} could not be found with: {v}")


def _get_structure(game_version: str, scenario_version: str) -> Dict[str, Any]:
    """
    Get the structure file of the given game and scenario version

    Args:
        game_version: The game version to get the structure file for
        scenario_version: The scenario version to get the structure file for

    Returns:
        The dictionary representation of the scenario for the specified versions

    Raises:
        InvalidScenarioStructureError: If the FileHeader section is not present in the loaded structure file
        UnknownScenarioStructureError: If the structure for the specified game/scenario version is not found
    """
    try:
        vdir = _get_version_directory_path()
        with (vdir / game_version / f'v{scenario_version}' / 'structure.json').open() as structure_file:
            structure = json.load(structure_file)

            if "FileHeader" not in structure.keys():
                raise InvalidScenarioStructureError(f"First section in structure should always be FileHeader.")
            return structure
    except FileNotFoundError:  # Unsupported version
        v = f"{game_version}:{scenario_version}"
        raise UnknownScenarioStructureError(f"The version {v} is not supported by AoE2ScenarioParser. :(") from None



"""

Warning: IMPORT THE MODULE, NOT THE VALUES!
    To change these settings, import the module: `from AoE2ScenarioParser import settings`.
    Then you can change the settings like: `settings.RAISE_ERROR_ON_WARNING = True`

If you import the values directly like: `from ... import RAISE_ERROR_ON_WARNING`
_(not written out because it wouldn't work)_, change the value would overwrite the reference and not the value which
means the change wouldn't transfer to the `settings` module.

"""

NOTIFY_UNKNOWN_BYTES = True
"""Show a notification of extra bytes being available at the end of the file, so you can notify the maintainer"""

# Reading / Writing settings
PRINT_STATUS_UPDATES = True
"""If status updates of what is being read and written should be printed to console"""
ALLOW_OVERWRITING_SOURCE = False
"""Disable the error being raised when overwriting source scenario"""
ALLOW_DIRTY_RETRIEVER_OVERWRITE = False
"""If it is allowed to overwrite a retriever that is dirty (it has been changed manually)"""
SHOW_VARIANT_WARNINGS = True
"""If warnings about incorrect variants should be shown or not"""

# Charset settings
MAIN_CHARSET = "utf-8"
"""The charset used to decode the text in the scenario. If it fails, will try the settings.FALLBACK_CHARSET"""
FALLBACK_CHARSET = "latin-1"
"""The charset used to decode the text in the scenario when the settings.MAIN_CHARSET fails"""



from __future__ import annotations

from typing import NamedTuple,TYPE_CHECKING

if TYPE_CHECKING:
    from AoE2ScenarioParser.sections.aoe2_file_section import AoE2FileSection


class ConstructProgress(NamedTuple):
    """a docstring"""
    section: 'AoE2FileSection'
    done: int



from __future__ import annotations

from typing import List, TYPE_CHECKING, Any

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.helper.bytes_conversions import bytes_to_int
from AoE2ScenarioParser.exceptions.asp_exceptions import EndOfFileError
from AoE2ScenarioParser.helper.incremental_generator import IncrementalGenerator

if TYPE_CHECKING:
    from AoE2ScenarioParser.sections.retrievers.retriever import Retriever

attributes = ['on_refresh', 'on_construct', 'on_commit']


def vorl(retriever: 'Retriever', value: List[Any]) -> List[Any] | Any:
    """
    vorl: Value or List. Checks if a value in a retriever is meant to be a variable or list, and returns
    it in that form.

    Args:
        retriever: The retriever
        value: The value to convert to the correct form

    Returns:
        The given list with value or the value inside it
    """
    if retriever.datatype.repeat != 1:
        return value

    if retriever.is_list is not None:
        return value if retriever.is_list else value[0]

    # Fallback to length check
    if len(value) == 1:
        return value[0]

    return value


def retrieve_bytes(igenerator: IncrementalGenerator, retriever: 'Retriever') -> List[bytes]:
    """
    Get the bytes required to set the data in the given retriever. Bytes are retriever from an incremental generator.

    Args:
        igenerator: The generator to return the bytes from
        retriever: The retriever for which the bytes need to be retrieved

    Returns:
        The corresponding bytes in a list.
    """
    var_type, var_len = retriever.datatype.type_and_length
    retrieved_bytes = []

    try:
        for i in range(retriever.datatype.repeat):
            if var_type != "str":  # (Signed) ints, floats, chars, plain bytes etc.
                retrieved_bytes.append(igenerator.get_bytes(var_len))
            else:  # String, Stored as: (signed int (n), string (string_length = n))
                int_bytes = igenerator.get_bytes(var_len)
                string_length = bytes_to_int(int_bytes, signed=True)
                string_bytes = igenerator.get_bytes(string_length)
                retrieved_bytes.append(int_bytes + string_bytes)
    except EndOfFileError:
        # END_OF_FILE_MARK retriever should always go in here
        if is_end_of_file_mark(retriever):
            retriever.datatype.repeat = 0
            return []

    # If more bytes present in the file after END_OF_FILE_MARK
    handle_end_of_file_mark(igenerator, retriever)

    return retrieved_bytes


def is_end_of_file_mark(retriever: 'Retriever') -> bool:
    """
    Args:
        retriever: The retriever to check

    Returns:
        Returns `True` if the retriever is the __END_OF_FILE_MARK__ retriever otherwise `False`
    """
    return retriever.name == "__END_OF_FILE_MARK__"


def handle_end_of_file_mark(igenerator: IncrementalGenerator, retriever: 'Retriever') -> None:
    """
    Prints a message if the file has more bytes in it than expected by the structure

    Args:
        igenerator: The generator to check if more bytes are present
        retriever: The retriever to check if it's the end of file mark
    """
    if is_end_of_file_mark(retriever) and settings.NOTIFY_UNKNOWN_BYTES:
        igenerator.progress -= retriever.datatype.length
        retrieved_bytes = igenerator.get_remaining_bytes()
        print("\n\n" + "\n".join([
            "The file being read has more bytes than anticipated.",
            "Please notify me (MrKirby/KSneijders) about this message!",
            "This will help with understanding more parts of scenario files! Thanks in advance!",
            "You can contact me using:",
            "- Discord: MrKirby#5063",
            "- Github: https://github.com/KSneijders/AoE2ScenarioParser/issues",
            "",
            "You can disable this check (and thereby this message) using:",
            ">>> from AoE2ScenarioParser import settings",
            ">>> settings.NOTIFY_UNKNOWN_BYTES = False",
            "",
            "Please be so kind and include the map in question. Thanks again!",
            "",
            f"Extra data found in the file ({len(retrieved_bytes)} bytes):",
            f"\t'{retrieved_bytes}'"
        ]))
        retriever.datatype.repeat = 1



from __future__ import annotations

from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_group import RetrieverObjectLinkGroup


class MessageManager(AoE2Object):
    """Manager of the everything message tab related."""

    _link_list = [
        RetrieverObjectLinkGroup("Messages", group=[
            RetrieverObjectLink("instructions", link="ascii_instructions"),
            RetrieverObjectLink("hints", link="ascii_hints"),
            RetrieverObjectLink("victory", link="ascii_victory"),
            RetrieverObjectLink("loss", link="ascii_loss"),
            RetrieverObjectLink("history", link="ascii_history"),
            RetrieverObjectLink("scouts", link="ascii_scouts"),

            RetrieverObjectLink("instructions_string_table_id", link="instructions"),
            RetrieverObjectLink("hints_string_table_id", link="hints"),
            RetrieverObjectLink("victory_string_table_id", link="victory"),
            RetrieverObjectLink("loss_string_table_id", link="loss"),
            RetrieverObjectLink("history_string_table_id", link="history"),
            RetrieverObjectLink("scouts_string_table_id", link="scouts"),
        ]),
    ]

    def __init__(self,
                 instructions: str,
                 hints: str,
                 victory: str,
                 loss: str,
                 history: str,
                 scouts: str,
                 instructions_string_table_id: int,
                 hints_string_table_id: int,
                 victory_string_table_id: int,
                 loss_string_table_id: int,
                 history_string_table_id: int,
                 scouts_string_table_id: int,
                 **kwargs
                 ):
        super().__init__(**kwargs)

        self.instructions: str = instructions
        self.hints: str = hints
        self.victory: str = victory
        self.loss: str = loss
        self.history: str = history
        self.scouts: str = scouts
        self.instructions_string_table_id: int = instructions_string_table_id
        self.hints_string_table_id: int = hints_string_table_id
        self.victory_string_table_id: int = victory_string_table_id
        self.loss_string_table_id: int = loss_string_table_id
        self.history_string_table_id: int = history_string_table_id
        self.scouts_string_table_id: int = scouts_string_table_id



from __future__ import annotations

from typing import Any, List, Type, Dict, TYPE_CHECKING, Optional
from uuid import UUID

from AoE2ScenarioParser.scenarios.scenario_store import getters
from AoE2ScenarioParser.sections.retrievers.construct_progress import ConstructProgress

if TYPE_CHECKING:
    from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
    from AoE2ScenarioParser.sections.retrievers.retriever import Retriever
    from AoE2ScenarioParser.sections.aoe2_file_section import AoE2FileSection
    from AoE2ScenarioParser.sections.retrievers.retriever_object_link_group import RetrieverObjectLinkGroup


class RetrieverObjectLinkParent:
    def __init__(
            self,
            section_name: str = None,
            link: str = None,
    ):
        self.section_name: str = section_name
        self.link: str = link
        self.splitted_link: List[str] = link.split('.') if link else []

        self.parent: Optional['RetrieverObjectLinkGroup'] = None
        """A link to the parent which will be injected by the group if it is in any """

    def pull_from_link(
            self,
            uuid: UUID,
            number_hist: List[int],
            host_obj: Type['AoE2Object'] = None,
            progress: ConstructProgress = None
    ) -> Any:
        """
        Retrieve value based on link

        Args:
            uuid: The UUID of the current scenario
            number_hist: The history numbers
            host_obj: A reference to the host object class
            progress: Start retrieving the value from a different starting point than the scenario sections

        Returns:
            The value located at the location found through self.link
        """
        return self.get_from_link(False, uuid, progress, number_hist)

    def get_section(self, uuid: UUID = None, progress: ConstructProgress = None) -> 'AoE2FileSection':
        """
        Wrapper function to return the ``from_section`` if it is defined. If not, get a section based on the current
        ``self.section_name`` from the scenario sections dict.

        Args:
            uuid: The UUID of the current scenario
            progress: The section to get another section from

        Returns:
            The section given or one from the scenario
        """
        if progress is None:
            sections = getters.get_sections(uuid)
            return sections[self.section_name]
        return progress.section

    def get_from_link(
            self,
            return_retriever=True,
            uuid: UUID = None,
            progress: ConstructProgress = None,
            number_hist: List[int] = None,
    ) -> 'Retriever' | Any:
        """
        Retrieve a retriever or the value inside it from a given link. Starting point (where the link should start
        looking from) can be from the scenario sections (if ``from_section`` is left empty). If ``from_section`` is not
        empty, it'll start looking from there. ``from_section`` is only used when this function is called from
        a RetrieverObjectLinkGroup.

        Args:
            return_retriever: If this function should return the retriever or the data inside of it
            uuid: The UUID of the scenario
            progress: If defined, use the link to start from this section instead of the main scenario sections
            number_hist: The number history of this object

        Returns:
            The found retriever or the value inside it depending on `return_retriever` parameter
        """
        section = self.get_section(uuid, progress)

        starting_index = 0
        if progress is not None:
            starting_index = progress.done

        # for index, item in enumerate(self.splitted_link, starting_index):
        for index in range(starting_index, len(self.splitted_link)):
            item = self.splitted_link[index]

            # If looking for a retriever and the end of the loop is found, return it
            if return_retriever and index == len(self.splitted_link) - 1:
                return section.retriever_map[item]

            if item.endswith("]"):
                # item[:-11] removes "[__index__]" from the key
                section = getattr(section, item[:-11])[number_hist[index]]
            else:
                section = getattr(section, item)

        # Can only be reached if `return_retriever=False`
        return section

    def overwrite_unsupported_properties(self, class_reference: Type['AoE2Object'], uuid: UUID) -> bool:
        """
        Overwrites the properties of class_reference if they are not supported in the version of the scenario with the
        given uuid

        Args:
            class_reference: A reference to the class containing the retriever object link ('self') in its link list
                (Mostly the host object where this link_list belongs to)
            uuid: The UUID of the current scenario

        Returns:
            `True` if values are (or already have been) overwritten
        """
        return False

    def get_names(self) -> List[str]:
        """
        Returns:
            All names from this link (singular/group) inside of a list
        """
        raise NotImplementedError("This function has not been implemented in the subclass yet.")

    def pull(
            self,
            uuid: UUID,
            number_hist: List[int] = None,
            host_obj: Type['AoE2Object'] = None,
            progress: ConstructProgress = None
    ) -> Dict[str, Any]:
        """
        Should result in the values being pulled from AoE2FileSection and Retrievers into the corresponding descendants
        of AoE2Object.

        Args:
            uuid: The UUID of the current scenario
            number_hist: The history number list
            host_obj: The host object that belongs to the retriever links
            progress: If defined, use the link to start from this section instead of the main scenario sections

        Returns:
            A Dict with keys as object constructor params and values corresponding to said keys
        """
        raise NotImplementedError("This function has not been implemented in the subclass yet.")

    def push(self, uuid: UUID, host_obj: 'AoE2Object') -> None:
        """
        Should result in the values going from the descendants of AoE2Object and being pushed back into the
        corresponding Retrievers and AoE2FileSection

        Args:
            uuid: The UUID of the current scenario
            host_obj: The host object that belongs to the retriever links
        """
        raise NotImplementedError("This function has not been implemented in the subclass yet.")

    @staticmethod
    def format_link_string(link: str, history: List[int]) -> str:
        """
        Format a link string to include the index values from the history

        Args:
            link: The link string (i.e. `"trigger_data[__index__].effect_data[__index__]"`)
            history: The number history i.e. `[4, 1]`

        Returns:
            The formatted string. Taking the examples from params, this would become `"trigger_data[4].effect_data[1]"`.
        """
        for i in range(len(history)):
            link = link.replace('__index__', str(history[i]), 1)
        return link



from typing import Optional, Tuple

from AoE2ScenarioParser.datasets.terrains import TerrainId
from AoE2ScenarioParser.helper.helper import i_to_xy
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.scenarios.scenario_store import getters
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_group import RetrieverObjectLinkGroup


class TerrainTile(AoE2Object):
    """Object for handling a tile in the map."""

    _link_list = [
        RetrieverObjectLinkGroup("Map", "terrain_data[__index__]", group=[
            RetrieverObjectLink("terrain_id"),
            RetrieverObjectLink("elevation"),
            RetrieverObjectLink("layer"),
        ]),
        RetrieverObjectLink("_index", retrieve_history_number=0),
    ]

    def __init__(self, terrain_id: int = TerrainId.GRASS_1, elevation: int = 0, layer: int = -1, _index: int = - 1,
                 **kwargs):
        self.terrain_id: int = terrain_id
        self.elevation: int = elevation
        self.layer: int = layer
        self._index: int = _index
        self._xy: Optional[Tuple[int, int]] = None

        super().__init__(**kwargs)

    @property
    def x(self) -> int:
        """The X coordinate of this tile on the map"""
        return self.xy[0]

    @property
    def y(self) -> int:
        """The Y coordinate of this tile on the map"""
        return self.xy[1]

    @property
    def i(self) -> int:
        """The index of this tile on the map"""
        return self._index

    @property
    def xy(self) -> Tuple[int, int]:
        """
        The X,Y coordinate of this tile on the map

        Returns:
            A tuple containing two integers representing the XY coordinates
        """
        if not self._xy:
            self._xy = i_to_xy(self._index, getters.get_map_size(self._uuid))
        return self._xy

    def _reset_terrain_index(self, new_index: int):
        """Reset the current terrain index"""
        self._index = new_index
        self._xy = None



class AoE2ScenarioParserWarning(Warning):
    """
    The super class of all warnings in the AoE2ScenarioParser.
    Can be used to act on all AoE2ScenarioParser Warnings at the same time
    """


class PythonVersionWarning(AoE2ScenarioParserWarning):
    """Shown when the user's python version might be incompatible with the libraries version in the (near) future"""
    pass


class ByteDecodeWarning(AoE2ScenarioParserWarning):
    """Shown when a non-fatal decoding issue occurs"""
    pass


class UpdateDirtyWarning(AoE2ScenarioParserWarning):
    """Shown when a manually edited object (aka dirty) is being updated/overwritten by a library process"""
    pass


class UuidForcedUnlinkWarning(AoE2ScenarioParserWarning):
    """Shown when a UUID object is updated in a way that forces the object to not be synced anymore"""
    pass


class XsScriptNameUsageWarning(AoE2ScenarioParserWarning):
    """Shown when someone uses the script_name value """
    pass


class IncorrectArmorAttackUsageWarning(AoE2ScenarioParserWarning):
    """
    Shown when a user uses calls that modify an effect's quantity/armor/attack class values.
    This can be intentional if you know what you're doing, so it deserves its own class so it can be disabled.
    """
    pass


class IncorrectVariantWarning(AoE2ScenarioParserWarning):
    """Shown when a user writes a file with a variant that is not AoE2 or RoR"""
    pass



from __future__ import annotations

from typing import List

from AoE2ScenarioParser.exceptions.asp_exceptions import InvalidScenarioStructureError
from AoE2ScenarioParser.helper.pretty_format import pretty_format_list
from AoE2ScenarioParser.helper.string_manipulations import add_tabs


class DependencyTarget:
    """
    Provides the objects for targeting a specific retriever based on it's section name (or 'self') and the retriever
    name.
    """

    def __init__(self, targets: List[List[str]]):
        """
        Args:
            targets: a list of lists which specify the targets of the dependency. Each sublist consists of two elements,
                the first element is the name of the section in which the target retriever lives and the second element
                is the name of the retriever within that section
        """
        self.targets = targets

    @classmethod
    def instance_or_none(cls, target: Union[str, List[str]]) -> DependencyTarget | None:
        """
        Takes in a string or a list of strings indicating the target retriever(s)

        Args:
            target: a string or list of strings representing the target retriever(s). Each string is in the format
                'SectionName:RetrieverName'.

        Returns:
            A DependencyTarget instance

        Raises:
            InvalidScenarioStructureError: if the parameter 'target' is set to anything other than a string or a list
                of strings
        """
        if target is None:
            return None

        if type(target) is str:
            return cls([target.split(":")])
        elif type(target) is list:
            return cls([entry.split(':') for entry in target])
        else:
            raise InvalidScenarioStructureError("Target defined using unknown type. "
                                                "For single targets use a string, else use a list of strings")

    def __repr__(self) -> str:
        return f"[DependencyTarget] {add_tabs(pretty_format_list(self.targets).strip(), 1)}"



import sys

from AoE2ScenarioParser.exceptions.asp_warnings import PythonVersionWarning
from AoE2ScenarioParser.helper.printers import warn

_block_below = (3, 8)
_notify_below = (3, 8)
_py_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"


def python_version_check():
    if sys.version_info < (3, 0):
        raise Exception("AoE2ScenarioParser does not support Python 2")

    if sys.version_info < _block_below:
        raise Exception(f"\n\nAoE2ScenarioParser requires the use of Python {'.'.join(map(str, _block_below))}.\n"
                        f"You are currently using version {_py_version}. You can download a newer version from: \n"
                        f"\thttps://www.python.org/downloads/\n")
    elif sys.version_info < _notify_below:
        v = '.'.join(map(str, _notify_below))
        warn('\n'.join([
            f"AoE2ScenarioParser will update the minimum python requirements to Python {v} in the (near) future.",
            f"You are currently using version {_py_version}. You can download a newer version from:",
            "\thttps://www.python.org/downloads/",
            f"You can disable these warnings using:",
            "\timport warnings",
            "\tfrom AoE2ScenarioParser.helper.exceptions import PythonVersionWarning",
            "\twarnings.filterwarnings(action='ignore', category=PythonVersionWarning)",
        ]), category=PythonVersionWarning)



from enum import Enum


class DependencyAction(Enum):
    """This is an enum class containing the values for the different retriever dependency actions"""
    REFRESH = 0
    REFRESH_SELF = 1
    SET_REPEAT = 2
    SET_VALUE = 3



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Tuple, Any, List, TextIO

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.helper.helper import typename
from AoE2ScenarioParser.helper.printers import s_print
from AoE2ScenarioParser.helper.string_manipulations import add_tabs, q_str, add_suffix_chars
from AoE2ScenarioParser.sections.aoe2_file_section import AoE2FileSection
from AoE2ScenarioParser.sections.retrievers.retriever import Retriever

if TYPE_CHECKING:
    from AoE2ScenarioParser.scenarios.aoe2_scenario import AoE2Scenario


def write_difference_to_file(
        file: TextIO,
        path: List[str],
        reason: str,
        difference: Tuple[Any, Any],
        data: Tuple[Any, Any]
) -> None:
    """
    Write a certain difference between 2 scenarios to a file in the format below
    
    Args:
        file: The debug file to write to
        path: The path where the difference occurred
        reason: The difference type (e.g. different values, different types, different list length etc.)
        difference: A tuple containing the compared values
        data: A tuple containing the data values of the compared retrievers
    """
    val0_type, val1_type = typename(data[0]), typename(data[1])
    max_type_len = max(len(val0_type), len(val1_type))

    if isinstance(data[0], list):
        l0, l1 = len(data[0]), len(data[1])
        val0_str = f"{data[0][0:4]}" + (f" ... (4 / {l0})" if l0 - 4 > 0 else "")
        val1_str = f"{data[1][0:4]}" + (f" ... (4 / {l1})" if l1 - 4 > 0 else "")
    else:
        val0_str = q_str(data[0])
        val1_str = q_str(data[1])

    file.write('\n\n\n' + '\n'.join([
        ' > '.join(path),
        add_tabs(
            '\n'.join([
                f'{reason} ({q_str(difference[0])} vs {q_str(difference[1])})',
                f'Values:',
                add_tabs('\n'.join([
                    f'{add_suffix_chars(val0_type, " ", max_type_len)} | {val0_str}',
                    f'{add_suffix_chars(val1_type, " ", max_type_len)} | {val1_str}'
                ]), 1)
            ]), 1)
    ]))


def compare_retrievers(
        output_file: TextIO,
        obj1: Retriever | AoE2FileSection,
        obj2: Retriever | AoE2FileSection,
        path: List[str]
) -> None:
    """
    Compare retriever data and the type of said data with the corresponding retriever data of another scenario.
    Recursively checks retrievers when they contain sections otherwise the retriever's data is verified.
    If data is not the same as the other scenario, it'll be logged to a given TextIO object

    Args:
        output_file: The file to write the inconsistencies (differences) to
        obj1: The source Retriever or FileSection to compare
        obj2: The Retriever or FileSection to compare against
        path: The current path inside the Retrievers and FileSections for logging purposes when differences are found.
    """
    path = path.copy()

    # If it's a file section, go through all retrievers in its map
    if isinstance(obj1, AoE2FileSection):
        retr2_keys = list(obj2.retriever_map.keys())

        for retr_key in obj1.retriever_map.keys():
            retriever1 = obj1.retriever_map[retr_key]

            if retr_key in obj2.retriever_map:
                retriever2 = obj2.retriever_map[retr_key]
                compare_retrievers(output_file, retriever1, retriever2, path + [retr_key])

                retr2_keys.remove(retr_key)
            else:
                write_difference_to_file(
                    output_file, path + [retr_key], "TARGET MISSING RETRIEVER",
                    (typename(retriever1.data), None),
                    (retriever1.data, None)
                )

        for retr2_key in retr2_keys:
            retriever2 = obj2.retriever_map[retr2_key]
            write_difference_to_file(
                output_file, path + [retr2_key], "SOURCE MISSING RETRIEVER",
                (None, typename(retriever2.data)),
                (None, retriever2.data)
            )

        return

    # If it's a retriever, check its data and go through it if necessary (when it's a list)
    # Check if the data types are the same
    if typename(obj1.data) != typename(obj2.data):
        write_difference_to_file(
            output_file, path, "DIFFERENT TYPES",
            (typename(obj1.data), typename(obj2.data)),
            (obj1.data, obj2.data)
        )
    # Check if both are lists (is for both because of previous check)
    elif isinstance(obj1.data, list):
        # Check if the lists are equal in size
        if len(obj1.data) != len(obj2.data):
            write_difference_to_file(
                output_file, path, "DIFFERENT LENGTHS",
                (len(obj1.data), len(obj2.data)),
                (obj1.data, obj2.data),
            )
        elif len(obj1.data) > 0:
            # Assuming a list only contains a single data type, compare the first of both lists
            if typename(obj1.data[0]) != typename(obj2.data[0]):
                write_difference_to_file(
                    output_file, path, "DIFFERENT LIST CONTENT TYPES",
                    (typename(obj1.data[0]), typename(obj2.data[0])),
                    (obj1.data, obj2.data),
                )
            # If the list is a list of AoE2FileSection, go through those recursively
            # We can assume type & length are equal due to earlier checks
            elif isinstance(obj1.data[0], AoE2FileSection):
                for i in range(len(obj1.data)):
                    compare_retrievers(output_file, obj1.data[i], obj2.data[i], path + [obj1.name + f"[{i}]"])
            else:
                # Both lists are empty
                for i in range(len(obj1.data)):
                    if obj1.data[i] != obj2.data[i]:
                        write_difference_to_file(
                            output_file, path + [obj1.name + f"[{i}]"], "DIFFERENT LIST DATA VALUE",
                            (obj1.data[i], obj2.data[i]),
                            (obj1.data[i], obj2.data[i]),
                        )
    # If it's not a list, compare the data directly.
    # This SHOULD only be values like: str, int, float or Enum datasets entries etc. (primitives)
    elif obj1.data != obj2.data:
        write_difference_to_file(
            output_file, path, "DIFFERENT VALUES",
            (obj1.data, obj2.data),
            (obj1.data, obj2.data)
        )


def debug_compare(
        scenario: 'AoE2Scenario',
        other: 'AoE2Scenario',
        filename,
        commit: bool = False,
        *,
        allow_multiple_versions: bool=False,
) -> None:
    """
    Compare two scenario files and report the differences found

    Args:
        scenario: The initial scenario
        other: The scenario to compare it to
        filename: The debug file to write the differences to
        commit: If the scenarios need to commit their manager changes before comparing (Defaults to False)
        allow_multiple_versions:
    """
    if commit:
        for scn in scenario, other:
            if hasattr(scn, '_object_manager'):
                scn.commit()

    s_print(f"Searching for differences between scenarios...")
    if not allow_multiple_versions:
        if other.game_version != scenario.game_version or other.scenario_version != scenario.scenario_version:
            raise ValueError("Scenarios must be from the same game & have the same version.")

    with open(filename, 'w', encoding=settings.MAIN_CHARSET) as output_file:
        for section_key in other.sections.keys():
            file_section1 = scenario.sections[section_key]
            file_section2 = other.sections[section_key]
            compare_retrievers(output_file, file_section1, file_section2, [section_key])

    s_print(f"Searching for differences between scenarios finished successfully.", final=True)
    s_print(f"File successfully written to: '{filename}'", color="magenta", final=True)



import math
from typing import List, TYPE_CHECKING
from uuid import UUID

from AoE2ScenarioParser.scenarios.scenario_store import getters
from AoE2ScenarioParser.sections.dependencies.dependency_action import DependencyAction

if TYPE_CHECKING:
    from AoE2ScenarioParser.sections.retrievers.retriever import Retriever
    from AoE2ScenarioParser.sections.dependencies.retriever_dependency import RetrieverDependency
    from AoE2ScenarioParser.sections.aoe2_file_section import AoE2FileSection


def refresh_targets(retriever_event: 'RetrieverDependency', section: 'AoE2FileSection', uuid: UUID) -> None:
    """
    Calls the execute_refresh_action on every target in the target list of the given retriever

    Args:
        retriever_event: The RetrieverDependency object that contains the target retrievers for the dependency action
        section: The AoE2FileSection object containing the given retriever
        uuid: The universally unique identifier for the scenario containing this file section
    """
    for target in retriever_event.dependency_target.targets:
        selected_retriever = select_retriever(target, section, uuid)
        execute_refresh_action(selected_retriever, section, uuid)


def execute_refresh_action(retriever: 'Retriever', section: 'AoE2FileSection', uuid: UUID) -> None:
    """
    Calls handle_retriever_dependency with the same retriever section and host, but changes the state to refresh.

    Args:
        retriever: The retriever to refresh
        section: The AoE2FileSection object containing the given retriever
        uuid: The universally unique identifier for the scenario containing this file section
    """
    handle_retriever_dependency(retriever, "refresh", section, uuid)


def handle_retriever_dependency(
        retriever: 'Retriever',
        state: str,
        section: 'AoE2FileSection',
        uuid: UUID
) -> None:
    """
    Checks if the given retriever has a dependency action for the specified state. If it does, then that action is
    carried out in this function.

    Args:
        retriever: The retriever to check the dependency for
        state: The state for which the dependency needs to be checked
        section: The AoE2FileSection object containing the given retriever
        uuid: The universally unique identifier for the scenario containing this file section
    """
    on_x = f'on_{state}'
    if not hasattr(retriever, on_x):
        return

    retriever_event = getattr(retriever, on_x)  # construct, push or refresh

    action = retriever_event.dependency_action

    if action == DependencyAction.REFRESH_SELF:
        execute_refresh_action(retriever, section, uuid)
    elif action == DependencyAction.REFRESH:
        refresh_targets(retriever_event, section, uuid)
    elif action in [DependencyAction.SET_VALUE, DependencyAction.SET_REPEAT]:
        value = execute_dependency_eval(retriever_event, section, uuid)
        if action == DependencyAction.SET_VALUE:
            retriever.set_data(value, affect_dirty=False)
        elif action == DependencyAction.SET_REPEAT:
            retriever.datatype.repeat = value


def execute_dependency_eval(retriever_event: 'RetrieverDependency', section: 'AoE2FileSection', uuid: UUID) -> int:
    """
    Runs the python code specified in a retriever dependency eval

    Args:
        retriever_event: The RetrieverDependency object that contains the eval
        section: The AoE2FileSection object containing the given retriever
        uuid: The universally unique identifier for the scenario containing this file section

    Returns:
        The result of the eval code in the given retriever dependency object
    """
    eval_code = retriever_event.dependency_eval.eval_code
    eval_locals = retriever_event.dependency_eval.eval_locals
    targets = retriever_event.dependency_target.targets

    values = []
    for target in targets:
        values.append(select_retriever(target, section, uuid).data)

    for index, target in enumerate(targets):
        eval_locals[target[1]] = values[index]
    eval_locals['math'] = math

    return eval(eval_code, {}, eval_locals)


def select_retriever(target: List[str], section: 'AoE2FileSection', uuid: UUID) -> 'Retriever':
    """
    Returns the retriever being targeted by a dependency action

    Args:
        target: A list containing the section and the name of the retriever
        section: The AoE2FileSection object containing the given retriever
        uuid: The universally unique identifier of the scenario containing the given file section

    Returns:
        The given retriever of the specified section (first element of target)
    """
    if target[0] == "self":
        return section.retriever_map[target[1]]
    else:
        sections = getters.get_sections(uuid)
        return sections[target[0]].retriever_map[target[1]]



from typing import Type

from AoE2ScenarioParser.objects.data_objects.trigger import Trigger


class TriggerSelect:
    """
    Object used to select a trigger in many trigger related functions. For ease of use, the alias `TS` can be
    called. You can also use those in combination with the class methods (factory methods). Like so:

    Tip: Selecting a trigger by its ID doesn't require this object
        Most functions allow you to just input `4` into the trigger_select parameter (instead of: `TS.index(4)`)

    Examples:
        `TS.index(4)`  To select the trigger with index 4

        `TS.display(4)`  Trigger with display index 4

        `TS.trigger(trigger)`  Well... The trigger object given...
    """
    def __init__(self, trigger_index: int = None, display_index: int = None, trigger: Trigger = None):
        """
        Args:
            trigger_index: The index of the trigger. Starting from 0, based on creation time
            display_index: The display index of a trigger. Starting from 0, based on display order in the editor
            trigger: The trigger object itself.
        """
        self.trigger = trigger
        self.display_index = display_index
        self.trigger_index = trigger_index

    @classmethod
    def index(cls, index: int):
        return cls(trigger_index=index)

    @classmethod
    def display(cls, display_index: int):
        return cls(display_index=display_index)

    @classmethod
    def trigger(cls, trigger: Trigger):
        return cls(trigger=trigger)


TS: Type[TriggerSelect] = TriggerSelect
"""Alias for `TriggerSelect`"""



from __future__ import annotations

from typing import Type

from AoE2ScenarioParser.objects.managers.de.map_manager_de import MapManagerDE
from AoE2ScenarioParser.objects.managers.de.trigger_manager_de import TriggerManagerDE
from AoE2ScenarioParser.objects.managers.de.unit_manager_de import UnitManagerDE
from AoE2ScenarioParser.objects.managers.de.xs_manager_de import XsManagerDE
from AoE2ScenarioParser.objects.managers.message_manager import MessageManager
from AoE2ScenarioParser.objects.managers.player_manager import PlayerManager
from AoE2ScenarioParser.scenarios.aoe2_scenario import AoE2Scenario, S


class AoE2DEScenario(AoE2Scenario):
    """
    Used to represent a scenario with version >= 1.36 (DE). It is the main class that is exposed to the user of the API.
    """

    @property
    def trigger_manager(self) -> TriggerManagerDE:
        """The trigger manager of the scenario"""
        return self._object_manager.managers['Trigger']

    @property
    def unit_manager(self) -> UnitManagerDE:
        """The unit manager of the scenario"""
        return self._object_manager.managers['Unit']

    @property
    def map_manager(self) -> MapManagerDE:
        """The map manager of the scenario"""
        return self._object_manager.managers['Map']

    @property
    def xs_manager(self) -> XsManagerDE:
        """The XS manager of the scenario"""
        return self._object_manager.managers['Xs']

    @property
    def player_manager(self) -> PlayerManager:
        """The player manager of the scenario"""
        return self._object_manager.managers['Player']

    @property
    def message_manager(self) -> MessageManager:
        return self._object_manager.managers['Message']

    @classmethod
    def from_file(cls: Type[S], path: str, game_version: str = "DE", name: str = "") -> S:
        """
        Creates and returns an instance of the AoE2DEScenario class from the given scenario file

        Args:
            path: The path to the scenario file to create the object from
            game_version: The version of the game to create the object for
            name: The name given to this scenario (defaults to the filename without extension)

        Returns:
            An instance of the AoE2DEScenario class which is the object representation of the given scenario file
        """
        return super().from_file(path=path, game_version=game_version, name=name)



from __future__ import annotations

from typing import Dict

from AoE2ScenarioParser.helper.pretty_format import pretty_format_dict
from AoE2ScenarioParser.sections.retrievers.retriever import Retriever


class AoE2StructModel:
    """
    Multiple retrievers containing related data are grouped together in scenario files under structs. This class is used
    to recursively build all retrievers and sub-structures inside a structure.

    Note that this class is just a model that represents each possible structure in a scenario file, hence it does NOT
    actually contain any data. Rather it acts as a "mold" to copy from and create multiple of structures that are
    repeated many times in a scenario. Thus, these struct models are actually always the same for a given version of a
    scenario. The actual data is held in file sections which are unique to each scenario file read.
    """
    def __init__(self, name: str, retriever_map: Dict[str, Retriever], structs: Dict[str, AoE2StructModel]):
        """
        Args:
            name: The name of the structure being built
            retriever_map: A dict of retrievers that constitute this structure
            structs: A dict of sub-structures inside this structure
        """
        self.name = name
        self.retriever_map = retriever_map
        self.structs = structs

    @classmethod
    def from_structure(cls, name: str, structure: Dict[str, Dict]) -> AoE2StructModel:
        """
        Recursively constructs all retrievers and sub-structures inside a given structure from its dictionary
        representation.

        Args:
            name (str): The name of the structure being constructed
            structure (Dict[str, Dict]): A dictionary representation of the structure containing the information about
            substructures and retrievers

        Returns:
            An AoE2StructModel instance representing the given structure
        """
        retriever_map = {}
        log_all_retrievers = structure.get('log', False)
        for retriever_name, attr in structure.get('retrievers').items():
            if log_all_retrievers:
                attr['log'] = True
            retriever_map[retriever_name] = Retriever.from_structure(retriever_name, attr)
        structs = model_dict_from_structure(structure)

        return cls(name, retriever_map, structs)

    def __str__(self):
        return f"[AoE2StructModel] {self.name} -> retrievers: " + pretty_format_dict(self.retriever_map)


def model_dict_from_structure(structure: Dict[str, Dict]) -> Dict[str, AoE2StructModel]:
    """
    Constructs all the structures in a given file section

    Args:
        structure: The dictionary representation of the file section

    Returns:
        A dictionary containing all the structures (AoE2StructModel objects) as key value pairs
    """
    models = {}
    for name, attr in structure.get('structs', {}).items():
        # Create struct model
        models[name] = AoE2StructModel.from_structure(name, attr)
    return models



from typing import List, Optional

from AoE2ScenarioParser.helper.helper import mutually_exclusive
from AoE2ScenarioParser.objects.data_objects.trigger import Trigger
from AoE2ScenarioParser.objects.data_objects.variable import Variable
from AoE2ScenarioParser.objects.managers.trigger_manager import TriggerManager
from AoE2ScenarioParser.objects.support.uuid_list import UuidList
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_group import RetrieverObjectLinkGroup


class TriggerManagerDE(TriggerManager):
    """Manager of all DE trigger related features"""
    _link_list = [
        RetrieverObjectLinkGroup("Triggers", group=[
            RetrieverObjectLink("triggers", link="trigger_data", process_as_object=Trigger),
            RetrieverObjectLink("trigger_display_order", link="trigger_display_order_array"),
            RetrieverObjectLink("variables", link="variable_data", process_as_object=Variable),
        ])
    ]

    def __init__(self, triggers: List[Trigger], trigger_display_order: List[int], variables: List[Variable], **kwargs):
        super().__init__(triggers, trigger_display_order, **kwargs)

        self.variables: List[Variable] = variables

    @property
    def variables(self) -> List[Variable]:
        """All currently renamed variables"""
        return self._variables

    @variables.setter
    def variables(self, value: List[Variable]):
        self._variables = UuidList(self._uuid, value)

    def add_variable(self, name: str, variable_id: int = -1) -> Variable:
        """
        Adds a variable.

        Args:
            name: The name for the variable
            variable_id: The ID of the variable. If left empty lowest available value will be used

        Returns:
            The newly renamed Variable
        """
        list_of_var_ids = [var.variable_id for var in self.variables]
        if variable_id == -1:
            for i in range(256):
                if i not in list_of_var_ids:
                    variable_id = i
                    break
            if variable_id == -1:
                raise IndexError(f"No variable ID available. All in use? In use: ({list_of_var_ids}/256)")
        if not (0 <= variable_id <= 255):
            raise ValueError("Variable ID has to fall between 0 and 255 (incl).")
        if variable_id in list_of_var_ids:
            raise ValueError("Variable ID already in use.")

        new_variable = Variable(variable_id=variable_id, name=name, uuid=self._uuid)
        self.variables.append(new_variable)
        return new_variable

    def get_variable(self, variable_id: int = None, variable_name: str = None) -> Optional[Variable]:
        """
        Get a specific variable

        Args:
            variable_id: The ID of the variable you want
            variable_name: The name of the variable you want

        Returns:
            The `Variable` object or None if it couldn't be found
        """
        if not mutually_exclusive(variable_id is not None, variable_name is not None):
            raise ValueError("Select a variable using either the variable_id or variable_name parameters.")
        for variable in self.variables:
            if variable.variable_id == variable_id or variable.name == variable_name:
                return variable
        return None

    def get_summary_as_string(self) -> str:
        """
        Create a human-readable string showcasing a summary of the content of the manager.
        This includes all triggers and the amount of conditions and effects they hold and also all renamed variables.

        Returns:
            The created string
        """
        return_string = super().get_summary_as_string()

        return_string += "\nVariables Summary:\n"
        if len(self.variables) == 0:
            return_string += "\t<< No Variables >>"

        longest_variable_name = -1
        for variable in self.variables:
            longest_variable_name = max(longest_variable_name, len(variable.name))

        longest_variable_name += 3
        for index, variable in enumerate(self.variables):
            var_name = variable.name
            name_buffer = " " * (longest_variable_name - len(var_name))
            return_string += f"\t{var_name}{name_buffer}[Index: {variable.variable_id}]\n"

        return return_string

    def get_content_as_string(self) -> str:
        """
        Create a human-readable string showcasing all content of the manager.
        This includes all triggers and their conditions and effects and also all renamed variables.

        This is also the function that is called when doing: `print(trigger_manager)`

        Returns:
            The created string
        """
        return_string = super().get_content_as_string()

        return_string += "Variables:\n"

        if len(self.variables) == 0:
            return_string += "\t<<No Variables>>\n"

        for variable in self.variables:
            return_string += f"\t'{variable.name}' [Index: {variable.variable_id}] ({variable._uuid})\n"

        return return_string



from typing import TypedDict

import requests
import regex

from common import tab, train_case

DefaultValue = str
UseDesc = str


class ResDesc(TypedDict):
    name: str
    desc: str
    defaults: dict[DefaultValue, UseDesc]
    note: str


ResId = int
ResDescs = dict[ResId, ResDesc]


def gen_docstr(desc: ResDesc) -> str:
    docstr = [
        tab('"' * 3),
        f"- Purpose: {desc['desc']}",
    ]

    if len(desc['defaults']) > 0:
        docstr.extend([
            "",
            "- Defaults:",
            "",
        ])

        for val, use in desc['defaults'].items():
            docstr.append(
                tab(f"- {val}: {use.strip()}")
            )  # use has a space prefix

    if len((desc['note'])) > 0:
        note = desc['note']

        if "See also:" in desc['note']:
            note = note.replace(")\n\n", ")\n")

            repl = r"    - https://ugc.aoe2.rocks/general/resources/resources/\1"
            note = regex.sub(r"\[.+\]\(\.\/(.+)\)", repl, note)

        docstr.extend([
            "",
            f"- Note: {note}",
        ])

    docstr.append('"' * 3)
    return "\n    ".join(docstr)


def gen_res_class(resources: ResDescs) -> str:
    klass = [r"class Attribute(_DataSetIntEnums):"]

    for id_, desc in resources.items():
        klass.extend([
            tab(train_case(desc["name"]) + f" = {id_}"),
            gen_docstr(desc)
        ])

    return "\n".join(klass)


def main():
    res = requests.get(
        r"https://raw.githubusercontent.com/Divy1211/AoE2DE_UGC_Guide/main/docs/general/resources/res_desc.json")
    resources = res.json()

    with open("resources.py", "w") as file:
        file.write(gen_res_class(resources))


if __name__ == "__main__":
    main()



class AoE2ScenarioParserError(Exception):
    """
    The super class of all errors in the AoE2ScenarioParser.
    Can be used to act on all AoE2ScenarioParser Errors at the same time
    """


class InvalidScenarioStructureError(AoE2ScenarioParserError):
    pass


class UnknownScenarioStructureError(AoE2ScenarioParserError):
    pass


class UnknownStructureError(AoE2ScenarioParserError):
    pass


class EndOfFileError(AoE2ScenarioParserError):
    pass


class UnsupportedAttributeError(AoE2ScenarioParserError):
    pass


class WarningToError(AoE2ScenarioParserError):
    pass


class UnsupportedVersionError(AoE2ScenarioParserError):
    pass


class ScenarioWritingError(AoE2ScenarioParserError):
    pass


def type_error_message(value, include_hint=True):
    return f"Expected int, found: {value.__class__}. " + (f"Maybe you meant: '{value}.ID'?" if include_hint else "")



from __future__ import annotations

from enum import IntEnum
from typing import Any

from AoE2ScenarioParser.datasets import conditions
from AoE2ScenarioParser.datasets.conditions import ConditionId
from AoE2ScenarioParser.helper.helper import raise_if_not_int_subclass, validate_coords
from AoE2ScenarioParser.helper.string_manipulations import add_tabs
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.support.attr_presentation import transform_condition_attr_value
from AoE2ScenarioParser.objects.support.trigger_object import TriggerComponent
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_group import RetrieverObjectLinkGroup
from AoE2ScenarioParser.sections.retrievers.support import Support


class Condition(AoE2Object, TriggerComponent):
    """Object for handling a condition."""
    hidden_attribute = 'condition_type'

    _link_list = [
        RetrieverObjectLinkGroup("Triggers", "trigger_data[__index__].condition_data[__index__]", group=[
            RetrieverObjectLink("condition_type"),
            RetrieverObjectLink("quantity"),
            RetrieverObjectLink("attribute"),
            RetrieverObjectLink("unit_object"),
            RetrieverObjectLink("next_object"),
            RetrieverObjectLink("object_list"),
            RetrieverObjectLink("source_player"),
            RetrieverObjectLink("technology"),
            RetrieverObjectLink("timer"),
            RetrieverObjectLink("area_x1"),
            RetrieverObjectLink("area_y1"),
            RetrieverObjectLink("area_x2"),
            RetrieverObjectLink("area_y2"),
            RetrieverObjectLink("object_group"),
            RetrieverObjectLink("object_type"),
            RetrieverObjectLink("ai_signal"),
            RetrieverObjectLink("inverted"),
            RetrieverObjectLink("variable"),
            RetrieverObjectLink("comparison"),
            RetrieverObjectLink("target_player"),
            RetrieverObjectLink("unit_ai_action", support=Support(since=1.40)),
            RetrieverObjectLink("object_state", support=Support(since=1.42)),
            RetrieverObjectLink("timer_id", support=Support(since=1.46)),
            RetrieverObjectLink("victory_timer_type", support=Support(since=1.46)),
            RetrieverObjectLink("include_changeable_weapon_objects", support=Support(since=1.46)),
            RetrieverObjectLink("xs_function", support=Support(since=1.40)),
        ])
    ]

    def __init__(
            self,
            condition_type: int = None,
            quantity: int = None,
            attribute: int = None,
            unit_object: int = None,
            next_object: int = None,
            object_list: int = None,
            source_player: int | IntEnum = None,
            technology: int | IntEnum = None,
            timer: int = None,
            area_x1: int = None,
            area_y1: int = None,
            area_x2: int = None,
            area_y2: int = None,
            object_group: int = None,
            object_type: int = None,
            ai_signal: int = None,
            inverted: int = None,
            variable: int = None,
            comparison: int = None,
            target_player: int | IntEnum = None,
            unit_ai_action: int = None,
            object_state: int = None,
            timer_id: int = None,
            victory_timer_type: int = None,
            include_changeable_weapon_objects: int = None,
            xs_function: str = None,
            **kwargs
    ):
        raise_if_not_int_subclass([object_list, technology])
        area_x1, area_y1, area_x2, area_y2 = validate_coords(area_x1, area_y1, area_x2, area_y2)

        self.condition_type: int = condition_type
        self.quantity: int = quantity
        self.attribute: int = attribute
        self.unit_object: int = unit_object
        self.next_object: int = next_object
        self.object_list: int = object_list
        self.source_player: int = source_player
        self.technology: int = technology
        self.timer: int = timer
        self.area_x1: int = area_x1
        self.area_y1: int = area_y1
        self.area_x2: int = area_x2
        self.area_y2: int = area_y2
        self.object_group: int = object_group
        self.object_type: int = object_type
        self.ai_signal: int = ai_signal
        self.inverted: int = inverted
        self.variable: int = variable
        self.comparison: int = comparison
        self.target_player: int = target_player
        self.unit_ai_action: int = unit_ai_action
        self.object_state: int = object_state
        self.timer_id: int = timer_id
        self.victory_timer_type: int = victory_timer_type
        self.include_changeable_weapon_objects: int = include_changeable_weapon_objects
        self.xs_function: str = xs_function

        super().__init__(**kwargs)

    def _should_be_displayed(self, attr: str, val: Any) -> bool:
        # Include the only exception to the -1 == invalid rule
        if self.condition_type == ConditionId.DIFFICULTY_LEVEL and attr == 'quantity' and val == -1:
            return True

        return super()._should_be_displayed(attr, val)

    def get_content_as_string(self, include_condition_definition: bool = False) -> str:
        """
        Create a human-readable string showcasing all content of this condition.

        This is also the function that is called when doing: `print(condition)`

        Args:
            include_condition_definition: If the condition meta-data should be added by this function

        Returns:
            The created string
        """
        if self.condition_type not in conditions.attributes:
            attributes_list = conditions.empty_attributes
        else:
            attributes_list = conditions.attributes[self.condition_type]

        return_string = ""
        for attribute in attributes_list:
            val = getattr(self, attribute)
            if not self._should_be_displayed(attribute, val):
                continue

            value_string = transform_condition_attr_value(self.condition_type, attribute, val, self._uuid)
            return_string += f"{attribute}: {value_string}\n"

        if return_string == "":
            return "<< No Attributes >>\n"

        if include_condition_definition:
            return f"{conditions.condition_names[self.condition_type]}:\n{add_tabs(return_string, 1)}"

        return return_string

    def __str__(self):
        return f"[Condition] {self.get_content_as_string(include_condition_definition=True)}"



from __future__ import annotations

import pickle
from time import sleep
from typing import Dict, List, Any

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.exceptions.asp_warnings import UpdateDirtyWarning
from AoE2ScenarioParser.helper import bytes_parser, string_manipulations
from AoE2ScenarioParser.helper.bytes_conversions import parse_bytes_to_val, parse_val_to_bytes
from AoE2ScenarioParser.helper.list_functions import listify
from AoE2ScenarioParser.helper.pretty_format import pretty_format_list
from AoE2ScenarioParser.helper.printers import warn, s_print
from AoE2ScenarioParser.sections.dependencies.dependency_action import DependencyAction
from AoE2ScenarioParser.sections.dependencies.retriever_dependency import RetrieverDependency
from AoE2ScenarioParser.sections.retrievers.datatype import DataType


class Retriever:
    """
    A retriever forms the fundamental unit of data in a scenario.
    All retrievers contain only one single value related to a particular thing in the scenario.
    This class is used to correctly interpret and store all the different types of fundamental data that can be
    present in a scenario
    """
    __slots__ = [
        'on_construct',
        'on_commit',
        'on_refresh',
        'name',
        'datatype',
        'is_list',
        'log_value',
        '_data',
        'default_value',
        'is_dirty',
    ]

    on_construct: RetrieverDependency
    on_commit: RetrieverDependency
    on_refresh: RetrieverDependency

    def __init__(
            self,
            name: str,
            default_value: Any = None,
            datatype: DataType = DataType(),
            is_list: bool | None = None,
            log_value: bool = False,
            data: Any = None
    ):
        """
        Args:
            name: The name of the item/piece of information to retrieved. Has to be unique within the Section or Struct
            default_value: The default value of this retriever
            datatype: An instance of the DataType class, specifying the datatype of the retriever
            is_list: If this retriever data should be presented using a list. If None, it's unknown.
            log_value: This will log information about this retriever when (Default: False)
                1. the data linked to it changes,
                2. it is constructed or committed
        """
        self.name: str = name
        self.default_value: Any = default_value
        self.datatype: DataType = datatype
        self.is_list: bool = is_list
        self.log_value: bool = log_value
        self.is_dirty: bool = False
        self._data: Any = data

        if log_value:
            self.datatype.log_value = True
            self.datatype._debug_retriever_name = name

    def get_data_as_bytes(self) -> bytes:
        """
        Get the data of a retriever in byte form

        Returns:
            Bytes that consist of the data from the retriever
        """
        self.update_datatype_repeat()

        if self.data is not None and self.datatype.repeat != 0:
            result = []
            if self.datatype.type == "struct":
                if isinstance(self.data, list):
                    for struct in self.data:
                        result.append(struct.get_data_as_bytes())
                else:
                    result.append(self.data.get_data_as_bytes())
            else:
                try:
                    for value in listify(self.data):
                        result.append(parse_val_to_bytes(self, value))
                except Exception as e:
                    print("\nException occurred during non-struct list creation: ")
                    print(self)
                    print(pretty_format_list(result))
                    raise e

            joined_result = b''.join(result)
        else:
            joined_result = b''

        if self.log_value:
            print(f"{self.to_simple_string()} (Data: {self.data}) retrieved: {joined_result}")

        return joined_result

    def set_data_from_bytes(self, bytes_list: List[bytes]) -> None:
        """
        Sets the data of a retriever from the list of bytes provided.

        Args:
            bytes_list (List(bytes)): a list of bytes

        Raises:
            ValueError: if no bytes are provided or if the repeat value of the retriever is not equal to the length of
                the list of bytes provided
        """
        if self.datatype.repeat > 0 and len(bytes_list) == 0:
            raise ValueError("Unable to set bytes when no bytes are given")
        if self.datatype.repeat > 0 and self.datatype.repeat != len(bytes_list):
            raise ValueError(f"Unable to set bytes when bytes list isn't equal to repeat (repeat: {self.datatype.repeat} vs list: {len(bytes_list)})")

        result = [parse_bytes_to_val(self, entry_bytes) for entry_bytes in bytes_list]
        self.data = bytes_parser.vorl(self, result)

    def update_datatype_repeat(self) -> None:
        """Sets all the datatype repeat values to the lengths of their containing lists"""
        if type(self.data) is list:
            self.datatype.repeat = len(self.data)

    @property
    def data(self):
        """The data contained in the retriever"""
        return self._data

    @data.setter
    def data(self, value):
        self.set_data(value)

    def set_data(self, value, affect_dirty: bool = True) -> None:
        """
        Setter wrapper to be able to circumvent updating the `dirty` attribute when called internally.

        Args:
            value: The value to set the data to
            affect_dirty: If the `dirty` attribute should be affected (set to true) by updating the data
        """
        if self.is_dirty and not affect_dirty:
            if settings.ALLOW_DIRTY_RETRIEVER_OVERWRITE:
                warn(f"Attribute {self.name} was overwritten by a writing process.", category=UpdateDirtyWarning)
            else:
                return
    
        if self.log_value:
            old_value = self._data
            self.print_value_update(old_value, value)

        # If repeat is 0 and value being said is truthy (mainly not an empty list) set repeat to one (convenience)
        if self.datatype.repeat == 0 and value:
            self.datatype.repeat = 1

        if affect_dirty and self._data is not None:
            self.is_dirty = True
        self._data = value

    def set_data_to_default(self):
        """Sets the values of all the data in the retriever to their default values"""
        if self.datatype.type == "data":
            data = bytes.fromhex(self.default_value)
        elif type(self.default_value) is list:
            data = self.default_value.copy()
        else:
            data = self.default_value

        self.data = data

    def duplicate(self) -> Retriever:
        """
        Creates and returns a copy of the retriever object

        Returns:
            A Retriever instance
        """
        retriever = Retriever(
            name=self.name,
            default_value=self.default_value,
            datatype=self.datatype.duplicate(),
            is_list=self.is_list,
            log_value=self.log_value
        )
        for attr in ['on_refresh', 'on_construct', 'on_commit']:
            if hasattr(self, attr):
                setattr(retriever, attr, getattr(self, attr))
        return retriever

    @classmethod
    def from_structure(cls, name: str, structure: dict) -> Retriever:
        """
        Creates a Retriever object from the given name and structure

        Args:
            name: The name of the retriever to create
            structure: A dict representation of the specified retriever's properties

        Returns:
            A Retriever instance
        """

        datatype = DataType(var=structure.get('type'), repeat=structure.get('repeat', 1))
        retriever = cls(
            name=name,
            default_value=structure.get('default'),
            datatype=datatype,
            is_list=structure.get('is_list', None),
            log_value=structure.get('log', False)
        )
        # Go through dependencies if exist, else empty dict
        for dependency_name, properties in structure.get('dependencies', {}).items():
            if type(properties) is not list:
                r_dep = RetrieverDependency.from_structure(properties)
                _evaluate_is_list_attribute(retriever, r_dep)
                setattr(retriever, dependency_name, r_dep)
            else:
                dependency_list = []
                for dependency_struct in properties:
                    r_dep = RetrieverDependency.from_structure(dependency_struct)
                    _evaluate_is_list_attribute(retriever, r_dep)
                    dependency_list.append(r_dep)
                setattr(retriever, dependency_name, dependency_list)

        return retriever

    def print_value_update(self, old: Any, new: Any) -> None:
        """
        Prints when data is changed. Can also be called for when data is changed but the property doesn't fire.
        This happens when an array is adjusted in size by appending to it ([...] += [...]).

        Prints the retriever, its new value and its old value

        Args:
            old: The old value
            new: The new value
        """
        print(f"{self.to_simple_string()} >>> set to: {string_manipulations.q_str(new)} "
              f"(was: {string_manipulations.q_str(old)})")

    def get_short_str(self):
        """
        Get a shorter string representation of the retriever containing its name, datatype and data

        Returns:
            A shorter string representation of the retriever
        """
        if self.data is not None:
            if type(self.data) is list:
                data = pretty_format_list(self.data)
            else:
                data = string_manipulations.q_str(self.data)
        else:
            data = "<None>"
        return f"{self.name} ({self.datatype.to_simple_string()}): {data}"

    def to_simple_string(self):
        """
        Get a string representation of the retriever containing its name, datatype, and default values

        Returns:
            A string representation of the retriever
        """
        return f"[Retriever] {self.name}: {self.datatype} (Default: {string_manipulations.q_str(self.default_value)})"

    def __repr__(self):
        if type(self.data) is list:
            data = str(pretty_format_list(self.data))
        else:
            data = string_manipulations.q_str(self.data)
        return f"{self.to_simple_string()} >>> {data}"


def duplicate_retriever_map(retriever_map: Dict[str, Retriever]) -> Dict[str, Retriever]:
    """
    Copies and returns the given dict of retrievers. This method is preferred over copy() or deepcopy()
    because of speed. Yes.

    Args:
        retriever_map: A dictionary with strings as keys and Retrievers as their values

    Returns:
        A copied version of the given dictionary
    """
    return pickle.loads(pickle.dumps(retriever_map))


def reset_retriever_map(retriever_map: Dict[str, Retriever]) -> None:
    """
    Sets all the values of the retrievers in the map to their default values.

    Args:
        retriever_map: The retriever map with retrievers to reset
    """
    for retriever in retriever_map.values():
        retriever.set_data_to_default()


def _evaluate_is_list_attribute(retriever: Retriever, dependency: RetrieverDependency) -> None:
    """
    Sets the is_list attribute of the given retriever to true if the dependency action of the specified
    dependency is SET_REPEAT and if the is_list attribute is not already set

    Args:
        retriever: The retriever to change the is_list value for
        dependency: The dependency to check the action for
    """
    if dependency.dependency_action == DependencyAction.SET_REPEAT and retriever.is_list is None:
        retriever.is_list = True



from typing import Type, List, Callable, Dict, Any, Optional, TYPE_CHECKING
from uuid import UUID

from AoE2ScenarioParser.exceptions.asp_exceptions import UnsupportedAttributeError
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.scenarios.scenario_store import getters
from AoE2ScenarioParser.sections.aoe2_file_section import AoE2FileSection
from AoE2ScenarioParser.sections.aoe2_struct_model import AoE2StructModel
from AoE2ScenarioParser.sections.dependencies.dependency import handle_retriever_dependency
from AoE2ScenarioParser.sections.retrievers.construct_progress import ConstructProgress
from AoE2ScenarioParser.sections.retrievers.retriever import Retriever
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_parent import RetrieverObjectLinkParent
from AoE2ScenarioParser.sections.retrievers.support import Support

if TYPE_CHECKING:
    from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class RetrieverObjectLink(RetrieverObjectLinkParent):
    def __init__(
            self,
            variable_name: str,
            section_name: str = None,
            link: str = None,
            support: Support = None,
            process_as_object: Type['AoE2Object'] = None,
            retrieve_history_number: int = None,
            commit_callback: Callable = None,
            destination_object: Type['AoE2Object'] = None,
    ):
        if link is not None and retrieve_history_number is not None:
            raise ValueError("You must use 'link' OR 'retrieve_history_number' as parameter, not both.")

        super().__init__(section_name, link or variable_name)

        self.name: str = variable_name
        self.support: Support = support
        self.process_as_object: Type['AoE2Object'] = process_as_object
        self.retrieve_history_number: Optional[int] = retrieve_history_number
        self.commit_callback: Callable = commit_callback
        self.destination_object: Type['AoE2Object'] = destination_object

        self.disabled = False
        """
        Set to `True` if the given property (referenced by link) is not supported in the current scenario version.
        When True, the properties are overridden to raise an error when used.  
        """

    def get_names(self):
        return [self.name]

    def overwrite_unsupported_properties(self, class_reference: Type['AoE2Object'], uuid: UUID) -> bool:
        scenario_version = getters.get_scenario_version(uuid)

        # If support is None, that means that the thing is supported in every scenario version that the parser supports
        if self.support is None:
            return False

        # If already disabled, the property has already been set to raise an error so return True
        if self.disabled:
            return True

        # If the property is supported, return False
        if self.support.supports(scenario_version):
            return False

        # Disable the property
        error_msg = self.get_unsupported_string(scenario_version)

        def _get(self_):
            raise UnsupportedAttributeError(error_msg)

        def _set(self_, val):
            if val is not None:
                raise UnsupportedAttributeError(error_msg)

        # Overwrite the destination object to disable when this object is just a mock object
        if self.destination_object is not None:
            class_reference = self.destination_object

        # Todo: Doesn't work properly when reading an older scenario first, and a newer one later
        #  Properties don't get reset!
        #  Just resetting doesnt work either as both scenarios can be used at the same time. get/set functions need
        #  more logic ("or Dynamic classes" -- Alian)
        setattr(class_reference, self.name, property(_get, _set))
        self.disabled = True
        return True

    def pull_from_link(
            self,
            uuid: UUID = None,
            number_hist: List[int] = None,
            host_obj: Type['AoE2Object'] = None,
            progress: ConstructProgress = None
    ) -> Any:
        overwritten = self.overwrite_unsupported_properties(host_obj, uuid)
        if overwritten:
            # Unsupported property; So unable to get value as it doesn't exist in current version.
            return None

        value = super().pull_from_link(uuid, number_hist, host_obj, progress)

        if self.process_as_object:
            value = self.process_object_list(value, number_hist, uuid)
        return value

    def pull(
            self,
            uuid: UUID,
            number_hist: List[int] = None,
            host_obj: Type['AoE2Object'] = None,
            progress: ConstructProgress = None
    ) -> Dict[str, Any]:
        if number_hist is None:
            number_hist = []

        if self.retrieve_history_number is not None:
            value = number_hist[self.retrieve_history_number]
        else:
            value = self.pull_from_link(uuid, number_hist, host_obj, progress)

        return {self.name: value}

    def process_object_list(
            self,
            objects: List['AoE2FileSection'],
            instance_number_history: List[int],
            uuid: UUID
    ):
        length = len(instance_number_history) + 1

        return [
            self.process_as_object.construct(uuid, instance_number_history + [index], progress=ConstructProgress(section=section, done=length))
            for index, section in enumerate(objects)
        ]

    def push_to_link(
            self,
            uuid: UUID = None,
            number_hist: List[int] = None,
            host_obj: 'AoE2Object' = None,
            progress: ConstructProgress = None
    ):
        if self.retrieve_history_number is not None:
            return
        if self.support and not self.support.supports(getters.get_scenario_version(uuid)):
            return

        retriever = self.get_from_link(True, uuid, progress, number_hist)

        if retriever is None:
            raise ValueError("RetrieverObjectLink is unable to find retriever")

        value = getattr(host_obj, self.name)
        file_section = self.get_section(uuid, progress)

        if self.commit_callback:
            value = self.commit_callback(host_obj, self.name, value)

        if self.process_as_object:
            struct_model = file_section.find_struct_model_by_retriever(retriever)

            RetrieverObjectLink.update_retriever_length(retriever, struct_model, len(value), uuid)
            RetrieverObjectLink.commit_object_list(value, host_obj.instance_number_history)
        else:
            retriever.set_data(value, affect_dirty=False)

        if hasattr(retriever, 'on_commit'):
            handle_retriever_dependency(retriever, "commit", file_section, uuid)

    def push(self, uuid: UUID, host_obj: 'AoE2Object') -> None:
        self.push_to_link(uuid, host_obj.instance_number_history, host_obj)

    @staticmethod
    def commit_object_list(object_list: List['AoE2Object'], instance_number_history: List[int]):
        for index, obj in enumerate(object_list):
            obj._instance_number_history = instance_number_history + [index]
            obj.commit()

    # Todo: Should this be a function in some other countries?
    @staticmethod
    def update_retriever_length(
            retriever: Retriever,
            model: AoE2StructModel,
            new_length: int,
            uuid: UUID
    ) -> None:
        """
        Update the length of a struct retriever. When committing the new data from the managers, certain lists like
        the trigger list might have changed in size. And, because every object (e.g. Trigger) corresponds to a section
        with the same data (e.g. Trigger Section) the 2 lists (in manager and the corresponding one in sections) should
        always be the same size. This function takes care of that.

        If the list is equal, leave it unchanged. If the list is shorter, cut the retriever data short. If the list is
        longer, add new sections to it based on the given model (& the model defaults)

        Args:
            retriever: The retriever containing the list
            model: The model inside the retriever, in case the list grew
            new_length: The new length of the list inside the managers
            uuid: The UUID of the current scenario
        """
        old_length = len(retriever.data)

        if new_length == old_length:
            return

        if new_length < old_length:
            retriever.set_data(retriever.data[:new_length], affect_dirty=False)
        elif new_length > old_length:
            retriever.data += [
                AoE2FileSection.from_model(model, uuid, set_defaults=True)
                for _ in range(new_length - old_length)
            ]

        if retriever.log_value:
            retriever.print_value_update(f"[{model.name}] * {old_length}", f"[{model.name}] * {new_length}")

    def get_unsupported_string(self, version: str):
        return f"The property '{self.name}' is {self.support}. Current version: {version}.\n"

    def __repr__(self):
        lines: List[str] = []

        if self.parent:
            lines.append(f"[RetrieverObjectLinkGroup] {self.parent.section_name}: {self.parent.link}")
            lines.append(f"\t[RetrieverObjectLink] {self.name}: {self.link} ({self.parent.link}.{self.link})")
        else:
            lines.append(f"[RetrieverObjectLink] {self.name}: {self.section_name}.{self.link}")

        if self.process_as_object is not None:
            lines.append(f"\t- Process as: {self.process_as_object.__name__}")

        if self.retrieve_history_number is not None:
            lines.append(f"\t- Get Hist Number: {self.retrieve_history_number}")

        if self.support is not None:
            lines.append(f"\t- Support: {self.support}")

        return '\n'.join([line for line in lines if line])



from __future__ import annotations

from copy import deepcopy
from enum import Enum
from typing import List, Any, Dict, TYPE_CHECKING, Optional
from uuid import UUID

from AoE2ScenarioParser.helper.pretty_format import pretty_format_dict
from AoE2ScenarioParser.helper.string_manipulations import add_tabs
from AoE2ScenarioParser.objects.support.uuid_list import NO_UUID
from AoE2ScenarioParser.scenarios.scenario_store import store
from AoE2ScenarioParser.sections.retrievers.construct_progress import ConstructProgress
from AoE2ScenarioParser.sections.retrievers.retriever_object_link_parent import RetrieverObjectLinkParent

if TYPE_CHECKING:
    from AoE2ScenarioParser.scenarios.aoe2_de_scenario import AoE2DEScenario


class AoE2Object:
    _link_list: List[RetrieverObjectLinkParent] = []

    def __init__(self, **kwargs):
        self._instance_number_history = []
        self._uuid = kwargs.get('uuid', NO_UUID)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            entry = self._deepcopy_entry(k, v)
            if entry is None:
                continue
            setattr(result, k, entry)
        return result

    def get_scenario(self) -> 'AoE2DEScenario':
        """ Get the scenario associated with the current object """
        scenario: Optional['AoE2DEScenario'] = store.get_scenario(uuid=self._uuid)
        if scenario:
            return scenario
        raise ValueError("Unable to fetch associated scenario from detached object")

    @property
    def instance_number_history(self):
        """
        Keeps indexes of the parents of this object. Should NOT be edited. Used for constructing/committing this object
        """
        return self._instance_number_history

    def _deepcopy_entry(self, k, v):
        if k in ['_sections', '_link_list']:
            val = getattr(self, k)
        else:
            val = deepcopy(v)
        return val

    @classmethod
    def construct(cls, uuid: UUID, number_hist: List[int] = None, progress: ConstructProgress = None):
        if number_hist is None:
            number_hist = []

        object_parameters: Dict[str, Any] = {}

        for link in cls._link_list:
            # print(f"\n\nCONSTRUCT [{link.link}]")
            # print(f"'{progress}' (progress (CONSTRUCT))")
            values = link.pull(uuid, number_hist, cls, progress)
            object_parameters.update(values)

        object_parameters['uuid'] = uuid

        return cls(**object_parameters)

    def commit(self, link_list: List[RetrieverObjectLinkParent] = None):
        """
        Commits all changes to the section & struct structure of the object it's called upon.

        Args:
            link_list: a separate list of RetrieverObjectLinks. This way it's possible to commit only specific
                properties instead of all from an object.
        """
        if link_list is None:
            link_list = self._link_list

        for link in reversed(link_list):
            link.push(self._uuid, host_obj=self)

    def _get_object_attrs(self):
        attrs = ["_instance_number_history", "_uuid"]
        for link in self._link_list:
            attrs.extend(link.get_names())
        return attrs

    def __repr__(self):
        attrs = self._get_object_attrs()

        self_dict = {}
        for attr in attrs:
            value = getattr(self, attr)
            if isinstance(value, Enum):
                self_dict[attr] = value.name
            else:
                self_dict[attr] = value

        return str(self.__class__.__name__) + ": " + add_tabs(pretty_format_dict(self_dict), 1)



from regex import regex


def tab(string: str) -> str:
    return f"    {string}"


def train_case(string: str) -> str:
    string = (
        string
        .upper()
        .replace(" ", "_")
        .replace("%", "PERCENT")
    )
    string = regex.sub(r"[!()/]", "", string)
    string = regex.sub(r"_+", "_", string)

    return string



from __future__ import annotations

from typing import Dict


class DependencyEval:
    """
    Provides the objects for storing dependency eval code and it's locals.
    """
    def __init__(self, eval_code: str, eval_locals: Dict | None = None):
        """
        Args:
            eval_code: The code executed using eval
            eval_locals: The locals dict handed to the eval function
        """
        if eval_locals is None:
            eval_locals = {}

        self.eval_code: str = eval_code
        self.eval_locals: Dict = eval_locals

    @classmethod
    def instance_or_none(cls, eval_code: str) -> DependencyEval | None:
        """
        Returns a DependencyEval instance created from the given code

        Args:
            eval_code (str): The code to create the DependencyEval object with

        Returns:
            An instance of the DependencyEval class
        """
        if eval_code is None:
            return None
        return cls(eval_code)



from __future__ import annotations

import struct
from typing import TYPE_CHECKING

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.exceptions.asp_warnings import ByteDecodeWarning
from AoE2ScenarioParser.helper.printers import warn
from AoE2ScenarioParser.helper.string_manipulations import has_str_trail, del_str_trail, q_str, \
    add_str_trail, trunc_string

if TYPE_CHECKING:
    from AoE2ScenarioParser.sections.retrievers.retriever import Retriever


def bytes_to_fixed_chars(byte_elements: bytes) -> str:
    """
    Converts the given bytes to characters using the specified codec

    Args:
        byte_elements: The bytes to convert

    Returns:
        A string of decoded characters
    """
    end_of_string = byte_elements.find(b"\x00")
    if end_of_string != -1:
        byte_elements = byte_elements[:end_of_string]
    return bytes_to_str(byte_elements)


def fixed_chars_to_bytes(string: str, var_len: int) -> bytes:
    """
    Converts the given string to bytes using the specified codec

    Args:
        string: The string to convert
        var_len: The length of the string. If shorter, appended with `b'\x00'` to fill up to length.

    Returns:
        bytes of the encoded string
    """
    bytes_ = str_to_bytes(string)
    return bytes_ + b"\x00" * (var_len - len(bytes_))


# @formatter:off
# Update tests if this changes
# Tests *should* fail - but just to be sure
_no_string_trail = [
    "filename",             # Scenario filename
    "ascii_instructions",   # Messages
    "ascii_hints",          # Messages
    "ascii_victory",        # Messages
    "ascii_loss",           # Messages
    "ascii_history",        # Messages
    "ascii_scouts",         # Messages
    "ascii_pregame",        # Cinematics
    "ascii_victory",        # Cinematics
    "ascii_loss",           # Cinematics
    "ascii_filename",       # Cinematics
    "strings",              # in PlayerDataTwo
    "ai_names",             # in PlayerDataTwo
    "ai_per_file_text",     # in AIStruct
    "unknown_string",       # in Map
    "map_color_mood",       # in Map
    "script_name",          # in Map
    "water_definition",     # in Map
    "message",              # in Trigger  [ADDED SO THEY'RE IGNORED - HANDLED IN EFFECT COMMIT CALLBACK]
    "sound_name",           # in Trigger  [ADDED SO THEY'RE IGNORED - HANDLED IN EFFECT COMMIT CALLBACK]
    "script_file_path",     # in Files
    "script_file_content",  # in Files
]
# @formatter:on


def bytes_to_str(
        byte_elements: bytes,
        charset: str = settings.MAIN_CHARSET,
        fallback_charset: str = settings.FALLBACK_CHARSET
) -> str | bytes:
    """
    Converts bytes to string based on given charset.

    Args:
        byte_elements: Bytes to be decoded to string.
        charset: Main charset used to decode the bytes. Defaults to settings.MAIN_CHARSET.
        fallback_charset: Fallback charset used to decode the bytes when the main fails.
            Defaults to settings.FALLBACK_CHARSET.

    Returns:
        The decoded string or the byte elements when the string cannot be decoded
    """
    trailless_elements = del_str_trail(byte_elements) if has_str_trail(byte_elements) else byte_elements

    for c in [charset, fallback_charset]:
        try:
            return trailless_elements.decode(c)
        except ValueError:
            continue

    # Return the string as bytes when it cannot be decoded. This will leave it as-is.
    truncated = trunc_string(byte_elements, 25)
    warn(f"Unable to decode bytes using '{charset}' and '{fallback_charset}', bytes: \n\t{truncated}",
         category=ByteDecodeWarning)
    return byte_elements


def str_to_bytes(
        string: str | bytes,
        charset: str = settings.MAIN_CHARSET,
        fallback_charset: str = settings.FALLBACK_CHARSET
) -> bytes:
    """
    Converts string to bytes based on given charsets

    Args:
        string: The string to convert
        charset: The main charset used while encoding
        fallback_charset: The fallback charset when the main fails

    Returns:
        The bytes converted from the given string

    Raises:
        ValueError: When the string cannot be encoded with either of the charsets
    """
    for c in [charset, fallback_charset]:
        try:
            return string.encode(c)
        except ValueError:
            continue
        except AttributeError as e:
            if type(string) is bytes:
                return string
            raise e
    raise ValueError(f"Unable to encode string using '{charset}' and '{fallback_charset}'. String:\n\t{q_str(string)}")


def bytes_to_int(byte_elements: bytes, endian: str = "little", signed: bool = True) -> int:
    """
    Converts the given bytes to an integer

    Args:
        byte_elements: The bytes to convert
        endian: If the conversion should be done with "little" or "big" endian. Defaults to "little"
        signed: If the int should be a signed int or not. Defaults to False

    Returns:
        The integer converted from the given bytes
    """
    return int.from_bytes(byte_elements, endian, signed=signed)


def int_to_bytes(integer: int, length: int, endian: str = "little", signed: bool = True) -> bytes:
    """
    Converts the given integer to bytes

    Args:
        integer: The integer to convert
        length: The number of bytes used to represent the integer
        endian: If the conversion should be done with "little" or "big" endian. Defaults to "little"
        signed: If the int is a signed int or not. Defaults to True

    Returns:
        The bytes converted from the given integer
    """
    return integer.to_bytes(length, byteorder=endian, signed=signed)


def bytes_to_float(byte_elements: bytes) -> float:
    """
    Converts the given bytes to a floating point value

    Args:
        byte_elements: The bytes to convert

    Returns:
        The floating point converted from the given bytes
    """
    return struct.unpack('f', byte_elements)[0]


def float_to_bytes(f: float) -> bytes:
    """
    Converts the given floating point value to bytes

    Args:
        f: The float to convert

    Returns:
        The bytes converted from the given float
    """
    return struct.pack('f', f)


def bytes_to_double(byte_elements: bytes) -> float:
    """
    Converts the given bytes to a double value

    Args:
        byte_elements: The bytes to convert

    Returns:
        The double value converted from the given bytes
    """
    return struct.unpack('d', byte_elements)[0]


def double_to_bytes(d: float) -> bytes:
    """
    Converts the given double value to bytes

    Args:
        d: The number to convert

    Returns:
        The bytes converted from the given double value
    """
    return struct.pack('d', d)


def parse_val_to_bytes(retriever: 'Retriever', val: int | float | str | bytes) -> bytes:
    """
    Encodes the given value to bytes according to the datatype of the retriever

    Args:
        retriever: The retriever to encode the bytes for
        val: The value to encode

    Returns:
        The encoded value as bytes

    Raises:
        ValueError: When the datatype of the retriever is not recognised or when the char string is longer than allowed.
    """
    var_type, var_len = retriever.datatype.type_and_length

    if var_type == "u" or var_type == "s":  # int
        return int_to_bytes(val, var_len, signed=(var_type == "s"))
    elif var_type == "str":  # str
        try:
            byte_string = str_to_bytes(val)
        except AttributeError as e:
            if type(val) is bytes:
                byte_string = val
            else:
                raise e from None

        if retriever.name not in _no_string_trail:
            byte_string = add_str_trail(byte_string)

        return _combine_int_str(byte_string, var_len, endian="little", signed=True, retriever=retriever)
    elif var_type == "c":  # str
        if len(val) > var_len:
            raise ValueError(f"Value cannot be longer than {var_len}, retriever: {retriever}")
        return fixed_chars_to_bytes(val, var_len)
    elif var_type == "data":  # bytes
        return val
    elif var_type == "f":  # float
        if var_len == 4:
            return float_to_bytes(val)
        else:
            return double_to_bytes(val)
    else:
        raise ValueError(f"Unable to parse value to bytes with unknown type: ({var_type})")


def parse_bytes_to_val(retriever: 'Retriever', byte_elements: bytes) -> int | float | str | bytes:
    """
    Decodes the bytes given according to the datatype of the retriever

    Args:
        retriever: The retriever to decode the bytes for
        byte_elements: The bytes to decode

    Returns:
        The interpreted value of the bytes

    Raises:
        ValueError: When the datatype of the retriever is not recognised
    """
    var_type, var_len = retriever.datatype.type_and_length

    if var_type == "u" or var_type == "s":
        return bytes_to_int(byte_elements, signed=(var_type == "s"))
    elif var_type == "str":
        return bytes_to_str(byte_elements[var_len:])
    elif var_type == "c":
        return bytes_to_fixed_chars(byte_elements)
    elif var_type == "data":
        return byte_elements
    elif var_type == "f":
        if var_len == 4:  # Float value
            return bytes_to_float(byte_elements)
        else:  # only 'else' is the trigger version. Which is a double (8 bytes)
            return bytes_to_double(byte_elements)
    else:
        raise ValueError(f"Unable to parse bytes with unknown type: ({var_type})")


def _combine_int_str(
        byte_string: bytes,
        length: int,
        endian: str,
        signed: bool,
        retriever: 'Retriever' = None
) -> bytes:
    """
    Appends the length of the given bytes (this length is in byte form itself) to the front of the bytes.
    This is done because strings are stored as an integer (as bytes) that is the length of the string followed by the
    same number of bytes representing that string

    Args:
        byte_string: The byte representation of a string
        length: The length of bytes to store the length in
        endian: This is the format used to encode the length to bytes
        signed: Determines if the length is a signed or unsigned integer
        retriever: The retriever to encode the bytes for

    Returns:
        the bytes of the length + the bytes of the string

    Raises:
        OverflowError: if the number of bytes needed to store the length is higher than the allowed value for scenarios
    """
    try:
        return int_to_bytes(len(byte_string), length, endian=endian, signed=signed) + byte_string
    except OverflowError as e:
        if str(e) == "int too big to convert":
            name: str = retriever.name if retriever else ""
            raise OverflowError(
                f"'{name.capitalize()}' is too long. "
                f"Length must be below {pow(2, length * 8 - 1)}, but is: {len(byte_string)}"
            ) from None



from __future__ import annotations

from AoE2ScenarioParser.sections.dependencies.dependency_action import DependencyAction
from AoE2ScenarioParser.sections.dependencies.dependency_eval import DependencyEval
from AoE2ScenarioParser.sections.dependencies.dependency_target import DependencyTarget


class RetrieverDependency:
    """
    Provides for objects used to handle dependencies between retrievers
    """
    def __init__(
            self,
            dependency_action: DependencyAction,
            dependency_target: DependencyTarget | None = None,
            dependency_eval: DependencyEval | None = None
    ):
        """
        Args:
            dependency_action: The type of action taken.
            dependency_target: The target of the action.
            dependency_eval: Eval code to execute for action.

        Raises:
            ValueError: if the parameter 'dependency_target' is None and the 'dependency_action' is anything other than
                REFRESH_SELF
        """
        if dependency_action != DependencyAction.REFRESH_SELF and dependency_target is None:
            raise ValueError(f"Parameter dependency_target cannot be None when action is {dependency_action.name}")

        self.dependency_action = dependency_action
        self.dependency_target = dependency_target

        if dependency_action in [DependencyAction.SET_REPEAT, DependencyAction.SET_VALUE] and dependency_eval is None:
            # Get name of first (and only) target, must define eval if using more than one target
            dependency_eval = DependencyEval(dependency_target.targets[0][1])

        self.dependency_eval = dependency_eval

    @classmethod
    def from_structure(cls, structure: dict) -> RetrieverDependency:
        """
        Creates a RetrieverDependency object from the given dependency structure.

        Args:
            structure: This is a dictionary that contains the action to perform, the target retriever to perform the
                action on and the relevant code to execute

        Returns:
            A RetrieverDependency instance created from the given dependency structure
        """
        return cls(
            dependency_action=DependencyAction[structure.get('action')],
            dependency_target=DependencyTarget.instance_or_none(structure.get('target')),
            dependency_eval=DependencyEval.instance_or_none(structure.get('eval')),
        )

    def __repr__(self):
        repr_string = f"\n[RetrieverDependency] {self.dependency_action}"
        if self.dependency_target:
            repr_string += f"\n\tTarget: {self.dependency_target}"
        if self.dependency_eval:
            repr_string += f"\n\tEval: {self.dependency_eval.eval_code}"
        return repr_string
